"""Test module for ecl2df.grid"""
import datetime
import os
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow
import pytest

from ecl2df import common, ecl2csv, grid
from ecl2df.eclfiles import EclFiles

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS")


def test_gridgeometry2df(mocker):
    """Test that dataframes are produced"""
    eclfiles = EclFiles(REEK)
    grid_geom = grid.gridgeometry2df(eclfiles)

    assert isinstance(grid_geom, pd.DataFrame)
    assert not grid_geom.empty

    assert "I" in grid_geom
    assert "J" in grid_geom
    assert "K" in grid_geom
    assert "X" in grid_geom
    assert "Y" in grid_geom
    assert "Z" in grid_geom
    assert "Z_MIN" in grid_geom
    assert "Z_MAX" in grid_geom
    assert "VOLUME" in grid_geom
    assert "ZONE" in grid_geom
    assert "GLOBAL_INDEX" in grid_geom

    # If at least one inactive cell, this will hold:
    assert grid_geom["GLOBAL_INDEX"].max() > len(grid_geom)

    assert (grid_geom["Z_MAX"] > grid_geom["Z_MIN"]).all()

    with pytest.raises(TypeError, match="missing 1 required positional"):
        grid.gridgeometry2df()

    with pytest.raises(AttributeError):
        # This error situation we don't really try to handle.
        grid.gridgeometry2df(None)

    with pytest.raises(ValueError, match="No EGRID file supplied"):
        mocker.patch("ecl2df.eclfiles.EclFiles.get_egridfile", return_value=None)
        grid.gridgeometry2df(eclfiles)


def test_wrongfile():
    """Test the EclFiles object on nonexistent files"""
    # pylint: disable=invalid-name,redefined-builtin

    # We can initalize this object with bogus:
    eclfiles = EclFiles("FOO.DATA")
    # but when we try to use it, things should fail:
    with pytest.raises(FileNotFoundError):
        grid.init2df(eclfiles)


def test_gridzonemap():
    """Check that zonemap can be merged automatically be default, and also
    that there is some API for supplying the zonemap directly as a dictionary"""
    eclfiles = EclFiles(EIGHTCELLS)
    grid_geom = grid.gridgeometry2df(eclfiles, zonemap=None)

    default_zonemap = grid_geom["ZONE"]

    grid_no_zone = grid.gridgeometry2df(eclfiles, zonemap={})
    assert "ZONE" not in grid_no_zone

    assert (grid.df(eclfiles, zonemap=None)["ZONE"] == default_zonemap).all()

    df_no_zone = grid.df(eclfiles, zonemap={})
    assert "ZONE" not in df_no_zone

    df_custom_zone = grid.gridgeometry2df(eclfiles, zonemap={1: "FIRSTLAYER"})
    assert "ZONE" in df_custom_zone
    assert set(df_custom_zone[df_custom_zone["K"] == 1]["ZONE"].unique()) == set(
        ["FIRSTLAYER"]
    )
    assert len(df_custom_zone) == len(grid_no_zone)

    df_bogus_zones = grid.gridgeometry2df(
        eclfiles, zonemap={999999: "nonexistinglayer"}
    )
    assert pd.isnull(df_bogus_zones["ZONE"]).all()

    # Test a custom "subzone" map via direct usage of merge_zone on an dataframe
    # where ZONE already exists:

    dframe = grid.df(eclfiles)
    subzonemap = {1: "SUBZONE1", 2: "SUBZONE2"}
    dframe = common.merge_zones(dframe, subzonemap, zoneheader="SUBZONE", kname="K")
    assert (dframe["ZONE"] == default_zonemap).all()
    assert set(dframe[dframe["K"] == 1]["SUBZONE"].unique()) == set(["SUBZONE1"])
    assert set(dframe[dframe["K"] == 2]["SUBZONE"].unique()) == set(["SUBZONE2"])
    assert len(dframe) == len(grid_no_zone)


def test_merge_initvectors():
    eclfiles = EclFiles(REEK)
    assert grid.merge_initvectors(eclfiles, pd.DataFrame(), []).empty
    foo_df = pd.DataFrame([{"FOO": 1}])
    pd.testing.assert_frame_equal(grid.merge_initvectors(eclfiles, foo_df, []), foo_df)

    with pytest.raises(ValueError, match="All of the columns"):
        grid.merge_initvectors(eclfiles, foo_df, ["NONEXISTING"])

    minimal_df = pd.DataFrame([{"I": 10, "J": 11, "K": 12}])

    with pytest.raises(KeyError):
        grid.merge_initvectors(eclfiles, minimal_df, ["NONEXISTING"])

    withporo = grid.merge_initvectors(eclfiles, minimal_df, ["PORO"])
    pd.testing.assert_frame_equal(
        withporo, minimal_df.assign(PORO=0.221848), check_dtype=False
    )

    with pytest.raises(ValueError):
        # ijknames must be length 3
        grid.merge_initvectors(
            eclfiles, minimal_df, ["PORO"], ijknames=["I", "J", "K", "L"]
        )
    with pytest.raises(ValueError):
        grid.merge_initvectors(eclfiles, minimal_df, ["PORO"], ijknames=["I", "J"])
    with pytest.raises(ValueError, match="All of the columns"):
        grid.merge_initvectors(eclfiles, minimal_df, ["PORO"], ijknames=["A", "B", "C"])


def test_init2df():
    """Test that dataframe with INIT vectors can be produced"""
    eclfiles = EclFiles(REEK)
    init_df = grid.init2df(eclfiles)

    assert isinstance(init_df, pd.DataFrame)
    assert not init_df.empty
    assert "PERMX" in init_df
    assert "PORO" in init_df
    assert "PORV" in init_df

    # The KRO data from the INIT file in Reek contains only NaN's,
    # but libecl gives out a large negative integer/float.
    # ecl2df should ensure this comes out as a NaN (but it
    # should be allowed later to drop columns which have only NaNs))
    if "KRO" in init_df:
        assert np.isnan(init_df["KRO"].unique()).all()


def test_grid_df():
    """Test that dataframe with INIT vectors and coordinates can be produced"""
    eclfiles = EclFiles(EIGHTCELLS)
    grid_df = grid.df(eclfiles)

    assert isinstance(grid_df, pd.DataFrame)
    assert not grid_df.empty
    assert "PERMX" in grid_df
    assert "PORO" in grid_df
    assert "PORV" in grid_df
    assert "I" in grid_df
    assert "J" in grid_df
    assert "K" in grid_df
    assert "X" in grid_df
    assert "Y" in grid_df
    assert "Z" in grid_df
    assert "VOLUME" in grid_df

    # Check that PORV is sensible
    assert (
        abs(sum(grid_df["PORO"] * grid_df["VOLUME"] - grid_df["PORV"]))
        / sum(grid_df["PORV"])
        < 0.00001
    )


def test_df2ecl(tmp_path):
    """Test if we are able to output include files for grid data"""
    eclfiles = EclFiles(REEK)
    grid_df = grid.df(eclfiles)

    fipnum_str = grid.df2ecl(grid_df, "FIPNUM", dtype=int)
    assert grid.df2ecl(grid_df, "FIPNUM", dtype="int", nocomments=True) == grid.df2ecl(
        grid_df, "FIPNUM", dtype=int, nocomments=True
    )
    with pytest.raises(ValueError, match="Wrong dtype argument foo"):
        grid.df2ecl(grid_df, "FIPNUM", dtype="foo")

    assert "FIPNUM" in fipnum_str
    assert "-- Output file printed by ecl2df.grid" in fipnum_str
    assert "35817 active cells" in fipnum_str  # (comment at the end)
    assert "35840 total cell count" in fipnum_str  # (comment at the end)
    assert len(fipnum_str) > 100

    fipnum_str_nocomment = grid.df2ecl(grid_df, "FIPNUM", dtype=int, nocomments=True)
    assert "--" not in fipnum_str_nocomment
    fipnum2_str = grid.df2ecl(
        grid_df, "FIPNUM", dtype=int, eclfiles=eclfiles, nocomments=True
    )
    # This would mean that we guessed the correct global size in the first run
    assert fipnum_str_nocomment == fipnum2_str

    float_fipnum_str = grid.df2ecl(grid_df, "FIPNUM", dtype=float)
    assert len(float_fipnum_str) > len(fipnum_str)  # lots of .0 in the string.

    fipsatnum_str = grid.df2ecl(grid_df, ["FIPNUM", "SATNUM"], dtype=int)
    assert "FIPNUM" in fipsatnum_str
    assert "SATNUM" in fipsatnum_str

    grid_df["FIPNUM"] = grid_df["FIPNUM"] * 3333
    fipnum_big_str = grid.df2ecl(grid_df, "FIPNUM", dtype=int)
    assert "3333" in fipnum_big_str
    assert len(fipnum_big_str) > len(fipnum_str)

    os.chdir(tmp_path)
    grid.df2ecl(grid_df, ["PERMX", "PERMY", "PERMZ"], dtype=float, filename="perm.inc")
    assert Path("perm.inc").is_file()
    incstring = open("perm.inc").readlines()
    assert sum([1 for line in incstring if "PERM" in line]) == 6

    assert grid.df2ecl(grid_df, ["PERMX"], dtype=float, nocomments=True) == grid.df2ecl(
        grid_df, ["PERMX"], dtype="float", nocomments=True
    )

    # with pytest.raises(ValueError, match="Wrong dtype argument"):
    grid.df2ecl(grid_df, ["PERMX"], dtype=dict)

    with pytest.raises(ValueError):
        grid.df2ecl(grid_df, ["PERMRR"])

    # Check when we have restart info included:
    gr_rst = grid.df(eclfiles, rstdates="all")
    fipnum_str_rst = grid.df2ecl(gr_rst, "FIPNUM", dtype=int, nocomments=True)
    assert fipnum_str_rst == fipnum_str_nocomment

    # When dates are stacked, there are NaN's  in the FIPNUM column,
    # which should be gracefully ignored.
    gr_rst_stacked = grid.df(eclfiles, rstdates="all", stackdates=True)
    fipnum_str_rst = grid.df2ecl(gr_rst_stacked, "FIPNUM", dtype=int, nocomments=True)
    assert fipnum_str_rst == fipnum_str_nocomment

    # dateinheaders here will be ignored due to stackdates:
    pd.testing.assert_frame_equal(
        gr_rst_stacked,
        grid.df(eclfiles, rstdates="all", stackdates=True, dateinheaders=True),
    )


def test_df2ecl_mock():
    """Test that we can use df2ecl for mocked minimal dataframes"""
    a_grid = pd.DataFrame(columns=["FIPNUM"], data=[[1], [2], [3]])
    simple_fipnum_inc = grid.df2ecl(
        a_grid, keywords="FIPNUM", dtype=int, nocomments=True
    )
    # (A warning is printed, that warning is warranted)
    assert "FIPNUM" in simple_fipnum_inc
    assert len(simple_fipnum_inc.replace("\n", " ").split()) == 5


def test_subvectors():
    """Test that we can ask for a few vectors only"""
    eclfiles = EclFiles(EIGHTCELLS)
    init_df = grid.init2df(eclfiles, "PORO")
    assert "PORO" in init_df
    assert "PERMX" not in init_df
    assert "PORV" not in init_df

    init_df = grid.init2df(eclfiles, "P*")
    assert "PORO" in init_df
    assert "PERMX" in init_df
    assert "PVTNUM" in init_df
    assert "SATNUM" not in init_df

    init_df = grid.init2df(eclfiles, ["P*"])
    assert "PORO" in init_df
    assert "PERMX" in init_df
    assert "PVTNUM" in init_df
    assert "SATNUM" not in init_df

    init_df = grid.init2df(eclfiles, ["P*", "*NUM"])
    assert "PORO" in init_df
    assert "PERMX" in init_df
    assert "PVTNUM" in init_df
    assert "SATNUM" in init_df
    assert "MULTZ" not in init_df


def test_dropconstants():
    """Test dropping of constants columns from dataframes"""
    df = pd.DataFrame(columns=["A", "B"], data=[[1, 1], [2, 1]])
    assert "B" not in grid.drop_constant_columns(df)
    assert "A" in grid.drop_constant_columns(df)
    assert "B" in grid.drop_constant_columns(df, alwayskeep="B")
    assert "B" in grid.drop_constant_columns(df, alwayskeep=["B"])

    with pytest.raises(TypeError):
        grid.drop_constant_columns(df, alwayskeep={"FOO"})
    with pytest.raises(TypeError):
        grid.drop_constant_columns(df, alwayskeep=1)
    with pytest.raises(TypeError):
        grid.drop_constant_columns({})

    assert grid.drop_constant_columns(pd.DataFrame()).empty


def test_df():
    """Test the df function"""
    eclfiles = EclFiles(REEK)
    # assert error..
    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        grid.df()

    grid_df = grid.df(eclfiles)
    assert not grid_df.empty
    assert "I" in grid_df  # From GRID
    assert "PORO" in grid_df  # From INIT
    assert "SOIL" not in grid_df  # We do not get RST unless we ask for it.

    grid_df = grid.df(eclfiles, vectors="*")
    assert "I" in grid_df  # From GRID
    assert "PORO" in grid_df  # From INIT
    assert "SOIL" not in grid_df  # We do not get RST unless we ask for it.

    grid_df = grid.df(eclfiles, vectors=["*"])
    assert "I" in grid_df  # From GRID
    assert "PORO" in grid_df  # From INIT
    assert "SOIL" not in grid_df  # We do not get RST unless we ask for it.

    grid_df = grid.df(eclfiles, vectors="PRESSURE")
    assert "I" in grid_df
    assert "PRESSURE" not in grid_df  # that vector is only in RST
    assert len(grid_df) == 35817
    assert "VOLUME" in grid_df

    grid_df = grid.df(eclfiles, vectors=["PRESSURE"])
    assert "I" in grid_df
    assert not grid_df.empty
    assert "PRESSURE" not in grid_df
    geometry_cols = len(grid_df.columns)

    grid_df = grid.df(eclfiles, vectors=["PRESSURE"], rstdates="last", stackdates=True)
    assert "PRESSURE" in grid_df
    assert len(grid_df.columns) == geometry_cols + 2
    assert "DATE" in grid_df  # Present because of stackdates

    grid_df = grid.df(eclfiles, vectors="PRESSURE", rstdates="last")
    assert "PRESSURE" in grid_df
    assert len(grid_df.columns) == geometry_cols + 1

    grid_df = grid.df(eclfiles, vectors="PRESSURE", rstdates="last", dateinheaders=True)
    assert "PRESSURE" not in grid_df
    assert "PRESSURE@2001-08-01" in grid_df

    grid_df = grid.df(
        eclfiles, vectors=["PORO", "PRESSURE"], rstdates="all", stackdates=True
    )
    assert "PRESSURE" in grid_df
    assert len(grid_df.columns) == geometry_cols + 3
    assert "DATE" in grid_df
    assert len(grid_df["DATE"].unique()) == 4
    assert not grid_df.isna().any().any()
    # Check that all but the dynamic data has been repeated:
    df1 = (
        grid_df[grid_df["DATE"] == "2000-01-01"]
        .drop(["DATE", "PRESSURE"], axis=1)
        .reset_index(drop=True)
    )
    df2 = (
        grid_df[grid_df["DATE"] == "2000-07-01"]
        .drop(["PRESSURE", "DATE"], axis=1)
        .reset_index(drop=True)
    )
    df3 = (
        grid_df[grid_df["DATE"] == "2001-02-01"]
        .drop(["PRESSURE", "DATE"], axis=1)
        .reset_index(drop=True)
    )
    df4 = (
        grid_df[grid_df["DATE"] == "2001-08-01"]
        .drop(["PRESSURE", "DATE"], axis=1)
        .reset_index(drop=True)
    )
    pd.testing.assert_frame_equal(df1, df2)
    pd.testing.assert_frame_equal(df1, df3)
    pd.testing.assert_frame_equal(df1, df4)

    grid_df = grid.df(eclfiles, vectors="PORO")
    assert "I" in grid_df
    assert "PORO" in grid_df
    assert len(grid_df) == 35817
    assert "DATE" not in grid_df

    grid_df = grid.df(eclfiles, vectors="PORO", rstdates="all")
    assert "I" in grid_df
    assert "PORO" in grid_df
    assert "DATE" not in grid_df
    # (no RST columns, so no DATE info in the dataframe)
    # (warnings should be printed)

    grid_df = grid.df(eclfiles, vectors="PORO", rstdates="all", stackdates=True)
    assert "I" in grid_df
    assert "PORO" in grid_df
    assert "DATE" not in grid_df
    # DATE is not included as it really does not make sense. The code
    # could have multiplied up the static dataframe each tagged with a date
    # but not for now.


def test_main(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "eclgrid.csv"
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "grid",
            EIGHTCELLS,
            "-o",
            str(tmpcsvfile),
            "--rstdates",
            "first",
            "--vectors",
            "PORO",
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty
    Path(tmpcsvfile).unlink()

    # Do again with also restarts and multiple vectors:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "grid",
            "--verbose",
            EIGHTCELLS,
            "-o",
            str(tmpcsvfile),
            "--rstdates",
            "2000-01-01",
            "--vectors",
            "PORO",
            "PERMX",
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty
    assert "PORO" in disk_df
    assert "PERMX" in disk_df
    Path(tmpcsvfile).unlink()

    # Test with constants dropping
    mocker.patch(
        "sys.argv", ["ecl2csv", "grid", REEK, "-o", str(tmpcsvfile), "--dropconstants"]
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    # That PVTNUM is constant is a particular feature
    # of the test dataset.
    assert "PVTNUM" not in disk_df
    assert not disk_df.empty


def test_main_arrow(tmp_path, mocker):
    """Check that we can export grid in arrow format"""
    mocker.patch(
        "sys.argv",
        ["ecl2csv", "grid", "--arrow", EIGHTCELLS, "-o", str(tmp_path / "grid.arrow")],
    )
    ecl2csv.main()

    # Obtain the CSV version for comparison:
    mocker.patch(
        "sys.argv", ["ecl2csv", "grid", EIGHTCELLS, "-o", str(tmp_path / "grid.csv")]
    )
    ecl2csv.main()

    # Read from disk and verify similarity
    disk_frame_arrow = pyarrow.feather.read_table(tmp_path / "grid.arrow").to_pandas()
    disk_frame_csv = pd.read_csv(tmp_path / "grid.csv")

    pd.testing.assert_frame_equal(disk_frame_arrow, disk_frame_csv, check_dtype=False)


def test_get_available_rst_dates():
    """Test the support of dates in restart files"""
    eclfiles = EclFiles(REEK)
    # rstfile = eclfiles.get_rstfile()

    alldates = grid.get_available_rst_dates(eclfiles)
    assert len(alldates) == 4

    didx = grid.dates2rstindices(eclfiles, "all")
    assert len(didx[0]) == len(alldates)
    assert len(didx[1]) == len(alldates)
    assert isinstance(didx[0][0], int)
    assert isinstance(didx[1][0], datetime.date)
    assert didx[1][0] == alldates[0]
    assert didx[1][-1] == alldates[-1]

    somedate = grid.dates2rstindices(eclfiles, "2000-07-01")
    assert somedate[1] == [alldates[1]]

    with pytest.raises(ValueError, match="date 1999-09-09 not found in UNRST file"):
        grid.dates2rstindices(eclfiles, "1999-09-09")

    with pytest.raises(ValueError, match="date 1999-0909 not understood"):
        grid.dates2rstindices(eclfiles, "1999-0909")

    expl_date = grid.dates2rstindices(eclfiles, datetime.date(2000, 7, 1))
    assert expl_date[1] == [alldates[1]]

    expl_datetime = grid.dates2rstindices(
        eclfiles, datetime.datetime(2000, 7, 1, 0, 0, 0)
    )
    assert expl_datetime[1] == [alldates[1]]

    expl_list_datetime = grid.dates2rstindices(eclfiles, [datetime.date(2000, 7, 1)])
    assert expl_list_datetime[1] == [alldates[1]]

    # For list input, only datetime.date objects are allowed:
    expl_list2_date = grid.dates2rstindices(
        eclfiles, [datetime.date(2000, 7, 1), datetime.date(2001, 2, 1)]
    )
    assert expl_list2_date[1] == [alldates[1], alldates[2]]

    with pytest.raises(ValueError, match="None of the requested dates were found"):
        grid.dates2rstindices(eclfiles, ["2000-07-01", "2001-02-01"])

    with pytest.raises(ValueError, match="None of the requested dates were found"):
        grid.dates2rstindices(
            eclfiles,
            [
                datetime.datetime(2000, 7, 1, 0, 0, 0),
                datetime.datetime(2001, 2, 1, 0, 0, 0),
            ],
        )

    with pytest.raises(ValueError, match="not understood"):
        grid.dates2rstindices(eclfiles, {"2000-07-01": "2001-02-01"})

    first = grid.dates2rstindices(eclfiles, "first")
    assert first[1][0] == alldates[0]

    last = grid.dates2rstindices(eclfiles, "last")
    assert last[1][0] == alldates[-1]

    dates = grid.get_available_rst_dates(eclfiles)
    assert isinstance(dates, list)

    # Test with missing RST file:
    eclfiles = EclFiles("BOGUS.DATA")
    with pytest.raises(IOError):
        eclfiles.get_rstfile()


def test_rst2df():
    """Test producing dataframes from restart files"""
    eclfiles = EclFiles(REEK)
    assert grid.rst2df(eclfiles, "first").shape == (35817, 24)
    assert grid.rst2df(eclfiles, "last").shape == (35817, 24)
    assert grid.rst2df(eclfiles, "all").shape == (35817, 23 * 4 + 1)

    assert "SOIL" in grid.rst2df(eclfiles, date="first", dateinheaders=False)
    assert (
        "SOIL@2000-01-01" in grid.rst2df(eclfiles, "first", dateinheaders=True).columns
    )

    rst_df = grid.rst2df(eclfiles, "first", stackdates=True)
    assert "DATE" in rst_df
    assert rst_df["DATE"].unique()[0] == "2000-01-01"
    rst_df = grid.rst2df(eclfiles, "all", stackdates=True)
    assert len(rst_df["DATE"].unique()) == len(grid.get_available_rst_dates(eclfiles))

    # "DATE" and "active" are now the extra columns:
    assert rst_df.shape == (4 * 35817, 23 + 2)

    # Test that only the PPCW column contains NaN's (only defined for selected cells)
    nancols = rst_df.isna().any()
    assert nancols["PPCW"]
    assert (
        len(rst_df[["PPCW", "DATE"]].dropna()["DATE"].unique()) == 4
    )  # All dates present
    assert sum(nancols) == 1  # All other columns are "False"

    # Check vector slicing:
    rst_df = grid.rst2df(eclfiles, "first", vectors="S???")
    assert rst_df.shape == (35817, 4)
    assert "SGAS" in rst_df
    assert "SWAT" in rst_df
    assert "SOIL" in rst_df  # This is actually computed
    assert "FIPWAT" not in rst_df

    rst_df = grid.rst2df(eclfiles, "first", vectors=["PRESSURE", "SWAT"])
    assert "PRESSURE" in rst_df
    assert "SWAT" in rst_df
    assert "SGAS" not in rst_df
    assert "SOIL" not in rst_df

    # Check that we can avoid getting SOIL if we are explicit:
    rst_df = grid.rst2df(eclfiles, "first", vectors=["SGAS", "SWAT"])
    assert "SOIL" not in rst_df
    assert "SGAS" in rst_df
    assert "SWAT" in rst_df



"""Parser and dataframe generator for the Eclipse keywords:
  * COMPDAT
  * COMPLUMP
  * COMPSEGS
  * WELOPEN
  * WELSEGS
  * WLIST
  * WSEGAICD
  * WSEGSICD
  * WSEGVALV
"""

import argparse
import datetime
import logging
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd

try:
    # pylint: disable=unused-import
    import opm.io.deck
except ImportError:
    # Allow parts of ecl2df to work without OPM:
    pass

from ecl2df import getLogger_ecl2csv

from .common import (
    get_wells_matching_template,
    merge_zones,
    parse_opmio_date_rec,
    parse_opmio_deckrecord,
    parse_opmio_tstep_rec,
    write_dframe_stdout_file,
)
from .eclfiles import EclFiles
from .grid import merge_initvectors

logger = logging.getLogger(__name__)

"""OPM authors and Roxar RMS authors have interpreted the Eclipse
documentation ever so slightly different when naming the data.

For COMPDAT dataframe columnnames, we prefer the RMS terms due to the
one very long one, and mixed-case in opm
"""
COMPDAT_RENAMER: Dict[str, str] = {
    "WELL": "WELL",
    "I": "I",
    "J": "J",
    "K1": "K1",
    "K2": "K2",
    "STATE": "OP/SH",
    "SAT_TABLE": "SATN",
    "CONNECTION_TRANSMISSIBILITY_FACTOR": "TRAN",
    "DIAMETER": "WBDIA",
    "Kh": "KH",
    "SKIN": "SKIN",
    "D_FACTOR": "DFACT",
    "DIR": "DIR",
    "PR": "PEQVR",
}

# Workaround an inconsistency in JSON-files for OPM-common < 2021.04:
WSEG_RENAMER: Dict[str, str] = {
    "SEG1": "SEGMENT1",
    "SEG2": "SEGMENT2",
}


def deck2dfs(
    deck: "opm.io.Deck",
    start_date: Optional[Union[str, datetime.date]] = None,
    unroll: bool = True,
) -> Dict[str, pd.DataFrame]:
    """Loop through the deck and pick up information found

    The loop over the deck is a state machine, as it has to pick up dates and
    potential information from the WELSPECS keyword.

    Args:
        deck: A deck representing the schedule
            Does not have to be a full Eclipse deck, an include file is sufficient
        start_date: The default date to use for
            events where the DATE or START keyword is not found in advance.
            Default: None
        unroll: Whether to unroll rows that cover a range,
            like K1 and K2 in COMPDAT and in WELSEGS. Defaults to True.

    Returns:
        Dictionary with dataframes, at least for COMPDAT, COMPSEGS and WELSEGS.
    """
    compdatrecords = []  # List of dicts of every line in input file
    compsegsrecords = []
    welopenrecords = []
    welsegsrecords = []
    wsegsicdrecords = []
    wsegaicdrecords = []
    wsegvalvrecords = []
    wlistrecords = []
    complumprecords = []
    welspecs = {}
    date = start_date  # DATE column will always be there, but can contain NaN/None
    for idx, kword in enumerate(deck):
        if kword.name in ["DATES", "START"]:
            for rec in kword:
                date = parse_opmio_date_rec(rec)
                logger.info("Parsing at date %s", str(date))
        elif kword.name == "TSTEP":
            if not date:
                logger.critical("Can't use TSTEP when there is no start_date")
                return {}
            for rec in kword:
                steplist = parse_opmio_tstep_rec(rec)
                # Assuming not LAB units, then the unit is days.
                days = sum(steplist)
                assert isinstance(date, datetime.date)
                date += datetime.timedelta(days=days)
                logger.info(
                    "Advancing %s days to %s through TSTEP", str(days), str(date)
                )
        elif kword.name == "WELSPECS":
            # Information from WELSPECS are to be used in case
            # 0 or 1* is used in the I or J column in COMPDAT
            # Only the latest information pr. well is stored.
            for wellrec in kword:
                welspecs_rec_dict = parse_opmio_deckrecord(wellrec, "WELSPECS")
                welspecs[welspecs_rec_dict["WELL"]] = {
                    "I": welspecs_rec_dict["HEAD_I"],
                    "J": welspecs_rec_dict["HEAD_J"],
                }
        elif kword.name == "COMPDAT":
            for rec in kword:  # Loop over the lines inside COMPDAT record
                rec_data = parse_opmio_deckrecord(
                    rec, "COMPDAT", renamer=COMPDAT_RENAMER
                )
                rec_data["DATE"] = date
                rec_data["KEYWORD_IDX"] = idx
                if rec_data["I"] == 0:
                    if rec_data["WELL"] not in welspecs:
                        raise ValueError(
                            "WELSPECS must be provided when I is defaulted in COMPDAT"
                        )
                    rec_data["I"] = welspecs[rec_data["WELL"]]["I"]
                if rec_data["J"] == 0:
                    if rec_data["WELL"] not in welspecs:
                        raise ValueError(
                            "WELSPECS must be provided when J is defaulted in COMPDAT"
                        )
                    rec_data["J"] = welspecs[rec_data["WELL"]]["J"]
                compdatrecords.append(rec_data)
        elif kword.name == "WSEGSICD":
            for rec in kword:  # Loop over the lines inside WSEGSICD record
                rec_data = parse_opmio_deckrecord(rec, "WSEGSICD", renamer=WSEG_RENAMER)
                rec_data["DATE"] = date
                rec_data["KEYWORD_IDX"] = idx
                wsegsicdrecords.append(rec_data)
        elif kword.name == "WSEGAICD":
            for rec in kword:  # Loop over the lines inside WSEGAICD record
                rec_data = parse_opmio_deckrecord(rec, "WSEGAICD", renamer=WSEG_RENAMER)
                rec_data["DATE"] = date
                rec_data["KEYWORD_IDX"] = idx
                wsegaicdrecords.append(rec_data)
        elif kword.name == "WSEGVALV":
            for rec in kword:  # Loop over the lines inside WSEGVALV record
                rec_data = parse_opmio_deckrecord(rec, "WSEGVALV")
                rec_data["DATE"] = date
                rec_data["KEYWORD_IDX"] = idx
                wsegvalvrecords.append(rec_data)
        elif kword.name == "COMPSEGS":
            wellname = parse_opmio_deckrecord(
                kword[0], "COMPSEGS", itemlistname="records", recordindex=0
            )["WELL"]
            for recidx in range(1, len(kword)):
                rec = kword[recidx]
                rec_data = parse_opmio_deckrecord(
                    rec, "COMPSEGS", itemlistname="records", recordindex=1
                )
                rec_data["WELL"] = wellname
                rec_data["DATE"] = date
                compsegsrecords.append(rec_data)
        elif kword.name == "WELOPEN":
            for rec in kword:
                rec_data = parse_opmio_deckrecord(rec, "WELOPEN")
                rec_data["DATE"] = date
                rec_data["KEYWORD_IDX"] = idx
                if rec_data["STATUS"] not in ["OPEN", "SHUT", "STOP", "AUTO", "POPN"]:
                    logger.warning(
                        (
                            "WELOPEN status %s is not a valid "
                            "COMPDAT state. Using 'SHUT' instead."
                        ),
                        rec_data["STATUS"],
                    )
                    rec_data["STATUS"] = "SHUT"
                welopenrecords.append(rec_data)
        elif kword.name == "WELSEGS":
            # First record contains meta-information for well
            # (opm deck returns default values for unspecified items.)
            welsegsdict = parse_opmio_deckrecord(
                kword[0], "WELSEGS", itemlistname="records", recordindex=0
            )
            # Loop over all subsequent records.
            for recidx in range(1, len(kword)):
                rec = kword[recidx]
                # WARNING: We assume that SEGMENT1 === SEGMENT2 (!!!) (if not,
                # we need to loop over a range just as for layer in compdat)
                rec_data = welsegsdict.copy()
                rec_data["DATE"] = date
                rec_data.update(
                    parse_opmio_deckrecord(
                        rec, "WELSEGS", itemlistname="records", recordindex=1
                    )
                )
                if "INFO_TYPE" in rec_data and rec_data["INFO_TYPE"] == "ABS":
                    rec_data["SEGMENT_MD"] = rec_data["SEGMENT_LENGTH"]
                welsegsrecords.append(rec_data)
        elif kword.name == "WLIST":
            for rec in kword:
                rec_data = parse_opmio_deckrecord(rec, "WLIST")
                rec_data["DATE"] = date
                if isinstance(rec_data["WELLS"], list):
                    rec_data["WELLS"] = " ".join(rec_data["WELLS"])

                # Do not store the asterisk that is needed in the Eclipse
                # keywords for referring to well lists:
                rec_data["NAME"] = rec_data["NAME"].replace("*", "")

                wlistrecords.append(rec_data)
        elif kword.name == "COMPLUMP":
            for rec in kword:  # Loop over the lines inside COMPLUMP record
                rec_data = parse_opmio_deckrecord(rec, "COMPLUMP")
                rec_data["DATE"] = date
                complumprecords.append(rec_data)

    compdat_df = pd.DataFrame(compdatrecords)
    welopen_df = pd.DataFrame(welopenrecords)
    wlist_df = pd.DataFrame(wlistrecords)
    complump_df = pd.DataFrame(complumprecords)

    if unroll and not compdat_df.empty:
        compdat_df = unrolldf(compdat_df, "K1", "K2")

    if not welopen_df.empty:
        compdat_df = applywelopen(
            compdat_df,
            expand_welopen(welopen_df, compdat_df),
            expand_wlist(wlist_df),
            unroll_complump(complump_df),
        )

    compsegs_df = pd.DataFrame(compsegsrecords)
    welsegs_df = pd.DataFrame(welsegsrecords)
    wsegsicd_df = pd.DataFrame(wsegsicdrecords)
    wsegaicd_df = pd.DataFrame(wsegaicdrecords)
    wsegvalv_df = pd.DataFrame(wsegvalvrecords)

    if unroll and not welsegs_df.empty:
        welsegs_df = unrolldf(welsegs_df, "SEGMENT1", "SEGMENT2")

    if unroll and not wsegsicd_df.empty:
        wsegsicd_df = unrolldf(wsegsicd_df, "SEGMENT1", "SEGMENT2")

    if unroll and not wsegaicd_df.empty:
        wsegaicd_df = unrolldf(wsegaicd_df, "SEGMENT1", "SEGMENT2")

    if "KEYWORD_IDX" in compdat_df.columns:
        compdat_df.drop(["KEYWORD_IDX"], axis=1, inplace=True)

    if "KEYWORD_IDX" in wsegsicd_df.columns:
        wsegsicd_df.drop(["KEYWORD_IDX"], axis=1, inplace=True)

    if "KEYWORD_IDX" in wsegaicd_df.columns:
        wsegaicd_df.drop(["KEYWORD_IDX"], axis=1, inplace=True)

    if "KEYWORD_IDX" in wsegvalv_df.columns:
        wsegvalv_df.drop(["KEYWORD_IDX"], axis=1, inplace=True)

    return dict(
        COMPDAT=compdat_df,
        COMPSEGS=compsegs_df,
        WELSEGS=welsegs_df,
        WELOPEN=welopen_df,
        WLIST=wlist_df,
        WSEGSICD=wsegsicd_df,
        WSEGAICD=wsegaicd_df,
        WSEGVALV=wsegvalv_df,
    )


def expand_welopen(welopen_df: pd.DataFrame, compdat_df: pd.DataFrame) -> pd.DataFrame:
    """Expands WELOPEN. First wildcard wells are expanded and then default
    coordinates are expanded. The order of the expansion is important.
    """
    exp_welopen_df = expand_welopen_wildcards(welopen_df, compdat_df)
    return expand_welopen_defaults(exp_welopen_df, compdat_df)


def expand_welopen_defaults(
    welopen_df: pd.DataFrame, compdat_df: pd.DataFrame
) -> pd.DataFrame:
    """Expands rows in WELOPEN where one or two coordinates are defaulted.
    Expansion happens by filtering on compdat rows that are matching the
    well name and the coordinates that are not defaulted.

    If all coordinates (I, J, K) are defaulted then the WELOPEN keyword
    is acting on the well and not on the connections, and shall not be
    expanded.

    It is important that expansion of wildcard wells is done prior to
    this function and that COMPDAT is unrolled so that K1=K2 in the input
    to this functions.
    """

    def is_default(value: Optional[int]) -> bool:
        if value is None or np.isnan(value):
            return True
        return value <= 0

    exp_welopen = []
    for _, row in welopen_df.iterrows():
        coord_defaulted = [is_default(row[coord]) for coord in ["I", "J", "K"]]
        if all(coord_defaulted) or not any(coord_defaulted):
            # If all coordinates are defaulted, the WELOPEN keyword is acting on
            # the well and not the connections, and shall not be expanded.
            # If no coordinates are defaulted, there is nothing to expand.
            exp_welopen.append(row)
        else:
            # If some of the coordinates are defaulted then we filter the
            # compdat dataframe to find the matching connections and expand
            # the WELOPEN row with those

            # Any compdat entry with DATE==None are kept as they
            # are assumed to have an earlier date than any dates defined
            compdat_filtered = compdat_df[compdat_df["DATE"].isnull()]

            # If the welopen entry DATE!=None we filter on compdat entries
            # <= this date
            if row["DATE"] is not None:
                compdat_filtered = pd.concat(
                    [compdat_filtered, compdat_df[compdat_df["DATE"] <= row["DATE"]]]
                )

            # Filter on well name
            compdat_filtered = compdat_filtered[compdat_filtered["WELL"] == row["WELL"]]

            # Filter on coordinates
            for coord in ["I", "J", "K"]:
                # In COMPDAT the K coordinate is named K1/K2.
                compdat_coord = coord + "1" if coord == "K" else coord
                if not is_default(row[coord]):
                    compdat_filtered = compdat_filtered[
                        compdat_filtered[compdat_coord] == row[coord]
                    ]

            # If compdat_filtered is empty it means that no connections are matching
            # the criterias in the WELOPEN row.
            if compdat_filtered.empty:
                raise ValueError(
                    "No connections are matching WELOPEN keyword with defaulted "
                    "coordinates:"
                    f"\n {row} "
                )

            for _, compdat_row in compdat_filtered.iterrows():
                exp_row = row.copy()
                exp_row["I"] = compdat_row["I"]
                exp_row["J"] = compdat_row["J"]
                exp_row["K"] = compdat_row["K1"]
                exp_welopen.append(exp_row)
    return pd.DataFrame(exp_welopen)


def expand_welopen_wildcards(
    welopen_df: pd.DataFrame, compdat_df: pd.DataFrame
) -> pd.DataFrame:
    """Expand rows in welopen with well names containing wildcard characters,
    with the correct wells from compdat_df that was defined at that date

    Example::

      WELOPEN
       'OP*' 'SHUT' /
      /

    might become the equivalent dataframe representation of::

      WELOPEN
       'OP1' 'SHUT' /
       'OP2' 'SHUT' /
      /

    if both OP1 and OP2 are defined at that time.

    Args:
        welopen_df: DataFrame with welopen
        compdat_df: DataFrame with compdat

    Returns:
        Expanded welopen dataframe
    """
    exp_welopen = []
    for _, row in welopen_df.iterrows():
        if row["WELL"].startswith("*"):
            # This means that the WELL field is refering to a well list
            # This is handled elsewhere and we let it pass here
            exp_welopen.append(row)
        elif "*" in row["WELL"] or "?" in row["WELL"]:
            # This row is identified a well name template with wildcard characters
            relevant_wells = compdat_df[compdat_df["DATE"] <= row["DATE"]][
                "WELL"
            ].unique()
            matched_wells = get_wells_matching_template(row["WELL"], relevant_wells)
            for well in matched_wells:
                exp_row = row.copy()
                exp_row["WELL"] = well
                exp_welopen.append(exp_row)
        else:
            exp_welopen.append(row)
    return pd.DataFrame(exp_welopen)


def unrolldf(
    dframe: pd.DataFrame, start_column: str = "K1", end_column: str = "K2"
) -> pd.DataFrame:
    """Unroll dataframes, where some column pairs indicate
    a range where data applies.

    After unrolling, column pairs with ranges are transformed
    into multiple rows, with no ranges.

    Example: COMPDAT supports K1, K2 intervals for multiple cells::

      COMPDAT
        'OP1' 33 44 10 11 /
      /

    is transformed/unrolled so it would be equal to::

      COMPDAT
        'OP1' 33 44 10 10 /
        'OP1' 33 44 11 11 /
      /

    The latter is easier to work with in Pandas dataframes

    Args:
        dframe: Dataframe to be unrolled
        start_column: Column name that contains the start of
            a range.
        end_column Column name that contains the corresponding end
            of the range.

    Returns:
        Dataframe, unrolled version. Identical to input if none of
        rows had any ranges.
    """
    if dframe.empty:
        return dframe
    if start_column not in dframe and end_column not in dframe:
        logger.warning(
            "Cannot unroll on non-existing columns %s and %s", start_column, end_column
        )
        return dframe
    start_eq_end_bools = dframe[start_column] == dframe[end_column]
    unrolled = dframe[start_eq_end_bools]
    list_unrolled = []
    if (~start_eq_end_bools).any():
        for _, rangerow in dframe[~start_eq_end_bools].iterrows():
            for k_idx in range(
                int(rangerow[start_column]), int(rangerow[end_column]) + 1
            ):
                rangerow[start_column] = k_idx
                rangerow[end_column] = k_idx
                list_unrolled.append(rangerow.copy())
    if list_unrolled:
        unrolled = pd.concat([unrolled, pd.DataFrame(list_unrolled)], axis=0)
    return unrolled


def unroll_complump(complump_df: pd.DataFrame) -> pd.DataFrame:
    """Unrolls the COMPLUMP keyword where K2>K1. Uses the unrolldf function,
    but this function gives more precise handling of errors.

    Example::

      COMPLUMP
       'OP1' 74 135 7 8 1 /
      /

    is transformed/unrolled to the dataframe representation of::

      COMPLUMP
       'OP1' 74 135 7 7 1 /
       'OP1' 74 135 8 8 1 /
      /

    Args:
        dframe: Dataframe to be unrolled

    Returns:
        Dataframe, unrolled version.
    """
    if complump_df is None or complump_df.empty:
        return complump_df

    for _, row in complump_df.iterrows():
        val_i = int(row["I"])
        val_j = int(row["J"])
        val_k1 = int(row["K1"])
        val_k2 = int(row["K2"])
        if val_i < 0 or val_j < 0 or val_k1 < 0 or val_k2 < 0:
            raise ValueError(
                f"Negative values for COMPLUMP coordinates are not allowed: {row}"
            )
        if val_i == 0 or val_j == 0 or val_k1 == 0 or val_k2 == 0:
            raise ValueError(
                f"Defaulted COMPLUMP coordinates are not supported in ecl2df: {row}"
            )
        if val_k2 < val_k1:
            raise ValueError(f"K2 must be equal to or greater than K1: {row}")
    return unrolldf(complump_df)


def expand_wlist(wlist_df: pd.DataFrame) -> pd.DataFrame:
    """Expand all WLIST actions in a dataframe into a dataframe with
    only "NEW" actions. This makes the dataframe cheaper to parse to
    get the state of the well lists at a particular date

    Example:

    .. code-block::

      WLIST
        '*OP' NEW OP1 /
        '*OP' ADD OP2 /
      /

    is transformed into the equivalent dataframe representation of:

    .. code-block::

      WLIST
        '*OP' NEW OP1 OP2 /  -- wells always sorted alphabetically
      /

    and then similarly for more complex MOV, DEL and NEW actions.

    The rationale is that if you extract all WLIST rows at a date
    and there are only NEW actions present, you can trust that dataframe
    to contain all WLIST actions accumulated from the start.

    Warning: If multiple WLIST keywords are at the same date, with other
    keywords depending on the WLIST state in between, that effect is lost
    through this dataframe representation.

    Args:
        wlist_df (pd.DataFrame): Dataframe with WLIST action rows. DATE
            must be present as a column in the DataFrame

    Returns:
        pd.DataFrame. WLIST rows with only NEW directives
    """

    # This function maintain all current (as in pr. date) well lists as a list
    # of dictionaries, which accumulates all WLIST directives. Every time the date
    # changes, the current state is outputted as it was valid for the previous date.

    currentstate: Dict[str, str] = {}

    if wlist_df.empty:
        return wlist_df

    currentdate = wlist_df["DATE"].min()
    new_records = []

    for _, wlist_record in wlist_df.iterrows():
        date = wlist_record["DATE"]
        if date > currentdate:
            # Store current state
            for wlistname, wells in currentstate.items():
                new_records.append(
                    {
                        "DATE": currentdate,
                        "NAME": wlistname,
                        "ACTION": "NEW",
                        "WELLS": wells,
                    }
                )
        currentdate = date

        if wlist_record["ACTION"] in ["ADD", "NEW"]:
            # Already defined well-lists can be used to append whole
            # well lists to other lists:
            recursive_wlists = [
                r_wlist
                for r_wlist in wlist_record["WELLS"].split()
                if r_wlist.startswith("*")
            ]
            for r_wlist in recursive_wlists:
                if r_wlist[1:] in currentstate:
                    wlist_record["WELLS"] = wlist_record["WELLS"].replace(
                        r_wlist, currentstate[r_wlist[1:]]
                    )
                else:
                    print(wlist_record)
                    raise ValueError(
                        f"Recursive well list {r_wlist} does not exist in "
                        f"{currentstate}"
                    )
        if wlist_record["ACTION"] == "NEW":
            currentstate[wlist_record["NAME"]] = " ".join(
                sorted(wlist_record["WELLS"].split())
            )
        elif wlist_record["ACTION"] in ["ADD", "DEL"]:
            if wlist_record["NAME"] not in currentstate:
                raise ValueError(
                    "WLIST ADD/DEL only works on existing well lists: "
                    f"{str(wlist_record)}"
                )
        if wlist_record["ACTION"] == "ADD":
            currentstate[wlist_record["NAME"]] = " ".join(
                sorted(
                    set(
                        wlist_record["WELLS"].split()
                        + currentstate[wlist_record["NAME"]].split()
                    )
                )
            )
        if wlist_record["ACTION"] == "DEL":
            currentstate[wlist_record["NAME"]] = " ".join(
                sorted(
                    list(
                        set(currentstate[wlist_record["NAME"]].split())
                        - set(wlist_record["WELLS"].split())
                    )
                )
            )
        if wlist_record["ACTION"] == "MOV":
            if wlist_record["NAME"] not in currentstate:
                currentstate[wlist_record["NAME"]] = ""
            currentstate[wlist_record["NAME"]] = " ".join(
                sorted(
                    list(
                        set(currentstate[wlist_record["NAME"]].split()).union(
                            set(wlist_record["WELLS"].split())
                        )
                    )
                )
            )
            for (
                wlist
            ) in currentstate.keys():  # pylint: disable=consider-iterating-dictionary
                if wlist == wlist_record["NAME"]:
                    continue
                currentstate[wlist] = " ".join(
                    sorted(
                        list(
                            set(currentstate[wlist].split())
                            - set(wlist_record["WELLS"].split())
                        )
                    )
                )

    # Dump final state:
    for wlistname, wells in currentstate.items():
        new_records.append(
            {"DATE": currentdate, "NAME": wlistname, "ACTION": "NEW", "WELLS": wells}
        )

    return pd.DataFrame(new_records)


def expand_complump_in_welopen_df(
    welopen_df: pd.DataFrame, complump_df: pd.DataFrame
) -> pd.DataFrame:
    """Go through all elements of WELOPEN and expand the rows
    referring to COMPLUMPS. The output dataframe is as if the
    connections were refered to explicitly in WELOPEN:

    Example: COMPLUMP and WELOPEN keywords::

      COMPLUMP
       'OP1' 74 135 7 7 1 /
       'OP1' 74 136 8 8 1 /
       'OP1' 74 136 9 9 2 /
       'OP1' 74 137 10 10 3 /
      /
      WELOPEN
       'OP1' 'SHUT' 3* 1 1 /
      /

    is transformed into the equivalent dataframe representation of::

      WELOPEN
       'OP1' 'SHUT' 74 135 7 /
       'OP1' 'SHUT' 74 135 8 /
      /

    Args:
        welopen_df: Dataframe with WELOPEN actions
        complump_df: Dataframe with COMPLUMP records. Optional.

    Returns:
        welopen_df with complump expanded to actual connections.

    """

    if (
        welopen_df is None
        or welopen_df.empty
        or complump_df is None
        or complump_df.empty
    ):
        return welopen_df

    exp_welopens = []
    for _, row in welopen_df.iterrows():
        if row["C1"] is None and row["C2"] is None:
            exp_welopens.append(row)
        elif row["C1"] is None or row["C2"] is None:
            raise ValueError(
                "Both or none of the completions numbers in "
                f"WELOPEN must be defined: {row}"
            )
        else:
            # Found a row that refers to complump numbers
            # Check that the cumplump numbers are ok:
            c_1, c_2 = int(row["C1"]), int(row["C2"])
            if c_1 < 0 or c_2 < 0:
                raise ValueError(f"Negative values for C1/C2 is not allowed: {row}")
            if c_1 == 0 or c_2 == 0:
                raise ValueError(f"Zeros for C1/C2 is not implemented: {row}")
            if c_2 < c_1:
                raise ValueError(f"C2 must be equal or greater than C1: {row}")

            relevant_complump_df = complump_df[
                (complump_df["DATE"] <= row["DATE"])
                & (complump_df["WELL"] == row["WELL"])
                & (complump_df["N"] >= c_1)
                & (complump_df["N"] <= c_2)
            ]
            # If I, J and K is also present in the WELOPEN record
            # we should "and" the complump filter and the i,j,k filter:
            if (
                {"I", "J", "K"}.issubset(row.keys())
                and row["I"]
                and row["J"]
                and row["K"]
            ):
                relevant_complump_df = relevant_complump_df[
                    (relevant_complump_df["I"] == row["I"])
                    & (relevant_complump_df["J"] == row["J"])
                    & (relevant_complump_df["K1"] == row["K"])
                    & (relevant_complump_df["K2"] == row["K"])
                ]
            for _, complump_row in relevant_complump_df.iterrows():
                if complump_row["K1"] != complump_row["K2"]:
                    raise ValueError(
                        "K1 must be equal to K2 in COMPLUMP "
                        f"(K2>K1 must be expanded elsewhere): {complump_row}"
                    )
                cell_row = row.copy()
                cell_row["I"] = complump_row["I"]
                cell_row["J"] = complump_row["J"]
                cell_row["K"] = complump_row["K1"]
                cell_row["C1"] = None
                cell_row["C2"] = None
                exp_welopens.append(cell_row)

    dframe = pd.DataFrame(exp_welopens)
    return dframe.astype(object).where(pd.notnull(dframe), None)


def expand_wlist_in_welopen_df(
    welopen_df: pd.DataFrame, wlist_df: pd.DataFrame
) -> pd.DataFrame:
    """Go trough a welopen dataframe and expand rows where the well-name refers
    to a well-list. The returned dataframe is as if the welopen commands were
    inputted explicitly without the use of WLIST.
    """
    if welopen_df.empty or welopen_df is None or wlist_df is None or wlist_df.empty:
        return welopen_df

    exp_welopens = []
    for _, row in welopen_df.iterrows():
        relevant_wlist_df = wlist_df[wlist_df["DATE"] <= row["DATE"]]
        if row["WELL"].startswith("*"):
            wlistname = row["WELL"].replace("*", "")
            if wlistname in relevant_wlist_df["NAME"].to_numpy():
                for well in (
                    relevant_wlist_df[relevant_wlist_df["NAME"] == wlistname]
                    .tail(1)["WELLS"]
                    .reset_index(drop=True)[0]
                    .split()
                ):
                    wellrow = row.copy()
                    wellrow.update({"WELL": well})
                    exp_welopens.append(wellrow)
            else:
                raise ValueError(f"Well list {wlistname} not defined at {row['DATE']}")
        else:
            # Explicit wellname was used, no expansion to happen:
            exp_welopens.append(row)
    dframe = pd.DataFrame(exp_welopens)
    return dframe.astype(object).where(pd.notnull(dframe), None)


def applywelopen(
    compdat_df: pd.DataFrame,
    welopen_df: pd.DataFrame,
    wlist_df: Optional[pd.DataFrame] = None,
    complump_df: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """Apply WELOPEN actions to the COMPDAT dataframe.

    Each record in the WELOPEN keyword acts as an operator on existing connections
    in existing wells.

    Example: COMPDAT and WELOPEN keyword::

      COMPDAT
       'OP1' 33 44 10 11 'OPEN' /
       'OP2' 66 44 10 11 'OPEN' /
      /
      WELOPEN
       'OP1' SHUT /
       'OP2' SHUT 66 44 10 /
      /

    This deck would define two wells where OP1 and OP2 have two connected grid cells
    each. The first welopen statment acts on the whole well, closing both the well and
    the connections. If this statement used STOP instead of SHUT, the connections would
    be left open. The second welopen statement acts on a single connection. Here SHUT
    and STOP would give the same result. This behavior has been proven to be correct
    in the simulator. The Eclipse manual states that 'If items 3 - 7 are all defaulted,
    the Open/Shut/Stop command applies to the well, leaving the connections unchanged',
    but this has been proven to be wrong. The state of the connection can be tested
    by looking at the CPI summary vectors. The connection is SHUT if CPI==0 and OPEN
    if CPI>0.

    WELOPEN can also be used at different dates and changes therefore the state of
    connections without explicit use of the COMPDAT keyword. This function translates
    WELOPEN actions into explicit additional COMPDAT definitions in the exported df.

    Args:
        compdat_df: Dataframe with unrolled COMPDAT data
        welopen_df: Dataframe with WELOPEN actions
        wlist_df: Dataframe with WLIST NEW records. Optional.
        complump_df: Dataframe with COMPLUMP records. Optional.

    Returns:
        compdat_df now including WELOPEN actions

    """
    if isinstance(wlist_df, pd.DataFrame):
        if wlist_df.empty:
            wlist_df = None
        else:
            if set(wlist_df["ACTION"]) != {"NEW"}:
                raise ValueError(
                    "The WLIST dataframe must be expanded through expand_wlist()"
                )

    welopen_df = welopen_df.astype(object).where(pd.notnull(welopen_df), None)
    welopen_df = expand_wlist_in_welopen_df(welopen_df, wlist_df)
    welopen_df = expand_complump_in_welopen_df(welopen_df, complump_df)

    for _, row in welopen_df.iterrows():
        acts_on_well = False
        if (row["I"] is None and row["J"] is None and row["K"] is None) or (
            row["I"] <= 0 and row["J"] <= 0 and row["K"] <= 0
        ):
            # Applies to all connections when the completion range
            # is set zero or negative.
            previous_state = compdat_df[
                (compdat_df["WELL"] == row["WELL"])
                & (compdat_df["KEYWORD_IDX"] < row["KEYWORD_IDX"])
            ].drop_duplicates(subset=["I", "J", "K1", "K2"], keep="last")
            acts_on_well = True
        elif (
            row["I"]
            and row["J"]
            and row["K"]
            and row["C1"] is None
            and row["C2"] is None
        ):
            # The compdat dataframe is assumed unrolled
            # so that K1 is always equal to K2. Any openings of lumped
            # connections (C1 and C2) should already be translated to
            # I, J, and K when we get here.
            previous_state = compdat_df[
                (compdat_df["WELL"] == row["WELL"])
                & (compdat_df["KEYWORD_IDX"] < row["KEYWORD_IDX"])
                & (compdat_df["I"] == row["I"])
                & (compdat_df["J"] == row["J"])
                & (compdat_df["K1"] == row["K"])
                & (compdat_df["K2"] == row["K"])
            ].drop_duplicates(subset=["I", "J", "K1", "K2"], keep="last")
        else:
            raise ValueError(
                "A WELOPEN keyword contains data that could not be parsed. "
                f"\n {row} "
            )

        if previous_state.empty:
            raise ValueError(
                "A WELOPEN keyword is not acting on any existing connection. "
                f"\n {row} "
            )

        new_state = previous_state

        # The COMPDAT DataFrame uses COMPDAT_RENAMER and therefore uses "OP/SH" as a
        # column name for the state of a well. WELOPEN uses "STATUS" for the state
        # column name and therefore a translation step needs to be done. The
        # underlying problem is that the opm-common definitions for the state of a
        # well in COMPDAT and WELOPEN are not identical. These translation steps can
        # be dropped when unity in the opm-common keyword definitions is reached.
        new_state["OP/SH"] = row["STATUS"].replace("POPN", "OPEN")

        # If the welopen statement acts on the whole well, then STOP closes the well
        # but opens the connections. If the welopen statement applies to selected
        # connections, then STOP means the same as SHUT.
        if acts_on_well:
            new_state["OP/SH"] = new_state["OP/SH"].replace("STOP", "OPEN")
        else:
            new_state["OP/SH"] = new_state["OP/SH"].replace("STOP", "SHUT")

        new_state["KEYWORD_IDX"] = row["KEYWORD_IDX"]
        new_state["DATE"] = row["DATE"]

        compdat_df = compdat_df.append(new_state)

    if not compdat_df.empty:
        compdat_df = (
            compdat_df.sort_values(by=["KEYWORD_IDX"])
            .drop_duplicates(subset=["WELL", "I", "J", "K1", "K2", "DATE"], keep="last")
            .reset_index(drop=True)
        )

    return compdat_df


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser: parser to fill with arguments
    """
    parser.add_argument("DATAFILE", help="Name of Eclipse DATA file.")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file.",
        default="compdat.csv",
    )
    parser.add_argument(
        "--initvectors",
        help="List of INIT vectors to merge into the data",
        nargs="+",
        default=None,
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def compdat_main(args):
    """Entry-point for module, for command line utility"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    compdat_df = df(eclfiles, initvectors=args.initvectors)
    write_dframe_stdout_file(compdat_df, args.output, index=False, caller_logger=logger)


def df(eclfiles: EclFiles, initvectors: Optional[List[str]] = None) -> pd.DataFrame:
    """Main function for Python API users

    Supports only COMPDAT information for now. Will
    add a zone-name if a zonefile is found alongside

    Returns:
        pd.Dataframe with one row pr cell to well connection
    """
    compdat_df = deck2dfs(eclfiles.get_ecldeck())["COMPDAT"]
    compdat_df = unrolldf(compdat_df)

    if initvectors:
        compdat_df = merge_initvectors(
            eclfiles, compdat_df, initvectors, ijknames=["I", "J", "K1"]
        )

    zonemap = eclfiles.get_zonemap()
    if zonemap:
        logger.info("Merging zonemap into compdat")
        compdat_df = merge_zones(compdat_df, zonemap)

    return compdat_df



import shutil
from pathlib import Path
from typing import Dict

import pytest

try:
    # pylint: disable=unused-import

    import ert_shared  # noqa
except ImportError:
    pytest.skip(
        "ERT is not installed. Skipping hook implementations.",
        allow_module_level=True,
    )

from ert_shared.plugins.plugin_manager import ErtPluginManager

import ecl2df.hook_implementations.jobs


@pytest.fixture(name="expected_jobs")
def fixture_expected_jobs(path_to_ecl2df: Path) -> Dict[str, Path]:
    """Dictionary of installed jobs with location to job configuration"""
    expected_job_names = [
        "ECL2CSV",
        "CSV2ECL",
    ]
    return {name: path_to_ecl2df / "config_jobs" / name for name in expected_job_names}


def test_hook_implementations(expected_jobs):
    """Test that the expected jobs can be found using an ERT plugin manager"""
    plugin_m = ErtPluginManager(plugins=[ecl2df.hook_implementations.jobs])

    installable_jobs = plugin_m.get_installable_jobs()
    for wf_name, wf_location in expected_jobs.items():
        assert wf_name in installable_jobs
        assert Path(installable_jobs[wf_name]).is_file()
        assert installable_jobs[wf_name].endswith(str(wf_location))

    assert set(installable_jobs.keys()) == set(expected_jobs.keys())

    expected_workflow_jobs = {}
    installable_workflow_jobs = plugin_m.get_installable_workflow_jobs()
    for wf_name, wf_location in expected_workflow_jobs.items():
        assert wf_name in installable_workflow_jobs
        assert installable_workflow_jobs[wf_name].endswith(wf_location)

    assert set(installable_workflow_jobs.keys()) == set(expected_workflow_jobs.keys())


def test_job_config_syntax(expected_jobs):
    """Check for syntax errors made in job configuration files"""
    for _, job_config in expected_jobs.items():
        # Check (loosely) that double-dashes are enclosed in quotes:
        for line in Path(job_config).read_text(encoding="utf8").splitlines():
            if not line.strip().startswith("--") and "--" in line:
                assert '"--' in line and " --" not in line


@pytest.mark.integration
def test_executables(expected_jobs):
    """Test executables listed in job configurations exist in $PATH"""
    for _, job_config in expected_jobs.items():
        executable = (
            Path(job_config).read_text(encoding="utf8").splitlines()[0].split()[1]
        )
        assert shutil.which(executable)


def test_hook_implementations_job_docs():
    """Test extracting docs from ERT hooks"""
    plugin_m = ErtPluginManager(plugins=[ecl2df.hook_implementations.jobs])

    installable_jobs = plugin_m.get_installable_jobs()

    docs = plugin_m.get_documentation_for_jobs()

    assert set(docs.keys()) == set(installable_jobs.keys())

    for job_name in installable_jobs.keys():
        print(job_name)
        assert docs[job_name]["description"] != ""
        assert docs[job_name]["category"] != "other"



"""Test module for nnc2df"""

import io
import subprocess
from pathlib import Path

import pandas as pd
import pytest

from ecl2df import ecl2csv, faults
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, nothing relevant in here then",
        allow_module_level=True,
    )

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS")


def test_faults2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(REEK)
    faultsdf = faults.df(eclfiles.get_ecldeck())

    assert "NAME" in faultsdf
    assert "I" in faultsdf
    assert "J" in faultsdf
    assert "K" in faultsdf
    assert "FACE" in faultsdf

    assert not faultsdf.empty


def test_str2df():
    """Test making dataframe from a string"""
    deckstr = """
FAULTS
  'A' 1 2 3 4 5 6 'I' /
  'B' 2 3 4 5 6 7 'J' /
/
"""
    deck = EclFiles.str2deck(deckstr)
    faultsdf = faults.df(deck)

    assert len(faultsdf) == 16


def test_nofaults():
    """Test on a dataset with no faults"""
    eclfiles = EclFiles(EIGHTCELLS)
    faultsdf = faults.df(eclfiles.get_ecldeck())
    assert faultsdf.empty


def test_multiplestr2df():
    """Test that we support multiple occurences of the FAULTS keyword"""
    deckstr = """
FAULTS
  'A' 1 2 3 4 5 6 'I' /
  'B' 2 3 4 5 6 7 'J' /
/
FAULTS
  'C' 1 1 3 3 10 15 'I' /
  'D' 2 2 4 4 10 10 'J' /
/
"""
    deck = EclFiles.str2deck(deckstr)
    faultsdf = faults.df(deck).set_index("NAME")

    assert len(faultsdf) == 23
    assert len(faultsdf.loc[["D"]]) == 1  # Pass lists to .loc for single row
    assert len(faultsdf.loc["C"]) == 6


def test_main_subparser(tmp_path, mocker):
    """Test command line interface with subparsers"""
    tmpcsvfile = tmp_path / "faultsdf.csv"
    mocker.patch("sys.argv", ["ecl2csv", "faults", REEK, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty


def test_magic_stdout():
    """Test that we can pipe the output into a dataframe"""
    result = subprocess.run(
        ["ecl2csv", "faults", "-o", "-", REEK], check=True, stdout=subprocess.PIPE
    )
    df_stdout = pd.read_csv(io.StringIO(result.stdout.decode()))
    assert not df_stdout.empty



"""
Extract EQUIL from an Eclipse deck as Pandas DataFrame
"""

import argparse
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

import pandas as pd

from ecl2df import common, getLogger_ecl2csv, inferdims

from .eclfiles import EclFiles

try:
    # pylint: disable=unused-import
    import opm.io

except ImportError:
    pass


logger = logging.getLogger(__name__)

SUPPORTED_KEYWORDS: List[str] = ["EQUIL", "PBVD", "PDVD", "RSVD", "RVVD"]
RENAMERS: Dict[str, Dict[str, Union[str, List[str]]]] = {}
RENAMERS["PBVD"] = {"DATA": ["Z", "PB"]}
RENAMERS["PDVD"] = {"DATA": ["Z", "PD"]}
RENAMERS["RSVD"] = {"DATA": ["Z", "RS"]}
RENAMERS["RVVD"] = {"DATA": ["Z", "RV"]}
RENAMERS["oil-water-gas"] = {
    "DATUM_DEPTH": "Z",
    "DATUM_PRESSURE": "PRESSURE",
    "OWC": "OWC",
    "PC_OWC": "PCOWC",
    "GOC": "GOC",
    "PC_GOC": "PCGOC",
    "BLACK_OIL_INIT": "INITRS",
    "BLACK_OIL_INIT_WG": "INITRV",
}
RENAMERS["gas-water"] = {
    "DATUM_DEPTH": "Z",
    "DATUM_PRESSURE": "PRESSURE",
    "OWC": "GWC",
    "PC_OWC": "PCGWC",
    "GOC": "IGNORE1",
    "PC_GOC": "IGNORE2",
    "BLACK_OIL_INIT": "IGNORE3",
    "BLACK_OIL_INIT_WG": "IGNORE4",
}
RENAMERS["oil-water"] = {
    "DATUM_DEPTH": "Z",
    "DATUM_PRESSURE": "PRESSURE",
    "OWC": "OWC",
    "PC_OWC": "PCOWC",
    "GOC": "IGNORE1",
    "PC_GOC": "IGNORE2",
    "BLACK_OIL_INIT": "IGNORE3",
    "BLACK_OIL_INIT_WG": "IGNORE4",
}
RENAMERS["oil-gas"] = {
    "DATUM_DEPTH": "Z",
    "DATUM_PRESSURE": "PRESSURE",
    "OWC": "IGNORE1",
    "PC_OWC": "IGNORE2",
    "GOC": "GOC",
    "PC_GOC": "PCGOC",
    "BLACK_OIL_INIT": "IGNORE3",
    "BLACK_OIL_INIT_WG": "IGNORE4",
}


def df(
    deck: Union[str, EclFiles, "opm.libopmcommon_python.Deck"],
    keywords: Optional[List[str]] = None,
    ntequl: Optional[int] = None,
) -> pd.DataFrame:
    """Extract EQUIL related keyword data, EQUIL, RSVD, RVVD
    PBVD and PDVD.

    How each data value in the EQUIL records are to be interpreted
    depends on the phase configuration in the deck, which means
    that we need more than the EQUIL section alone to determine the
    dataframe.

    If ntequl is not supplied and EQLDIMS is not in the deck, the
    equil data is not well defined in terms of OPM. This means
    that we have to infer the correct number of EQUIL lines from what
    gives us successful parsing from OPM. In those cases, the
    deck must be supplied as a string, if not, extra EQUIL lines
    are possibly already removed by the OPM parser in eclfiles.str2deck().

    Arguments:
        deck: Eclipse deck or string with deck. If
           not string, EQLDIMS must be present in the deck.
        keywords: Requested keywords for which to extract data.
        ntequl: If not None, should state the NTEQUL in EQLDIMS. If
            None and EQLDIMS is not present, it will be inferred.

    Return:
        pd.DataFrame, at least with columns KEYWORD and EQLNUM
    """
    if isinstance(deck, EclFiles):
        deck = deck.get_ecldeck()

    deck = inferdims.inject_xxxdims_ntxxx("EQLDIMS", "NTEQUL", deck, ntequl)
    ntequl = deck["EQLDIMS"][0][inferdims.DIMS_POS["NTEQUL"]].get_int(0)

    wanted_keywords = common.handle_wanted_keywords(keywords, deck, SUPPORTED_KEYWORDS)

    frames = []
    for keyword in wanted_keywords:
        # Construct the associated function names
        function_name = keyword.lower() + "_fromdeck"
        function = globals()[function_name]
        dframe = function(deck, ntequl=ntequl)
        frames.append(dframe.assign(KEYWORD=keyword))
    nonempty_frames = [frame for frame in frames if not frame.empty]
    if nonempty_frames:
        dframe = pd.concat(nonempty_frames, axis=0, sort=False, ignore_index=True)
        logger.info(
            "Extracted keywords %s for %g EQLNUMs",
            dframe["KEYWORD"].unique(),
            len(dframe["EQLNUM"].unique()),
        )
        return dframe
    logger.warning("No equil data found")
    return pd.DataFrame()


def rsvd_fromdeck(
    deck: Union[str, "opm.libopmcommon_python.Deck"], ntequl: Optional[int] = None
) -> pd.DataFrame:
    """Extract RSVD data from a deck

    Args:
        deck
        ntequl: Number of EQLNUM regions in deck. Will
            be inferred if not present in deck
    """
    if "EQLDIMS" not in deck:
        deck = inferdims.inject_xxxdims_ntxxx("EQLDIMS", "NTEQUL", deck, ntequl)
    return common.ecl_keyworddata_to_df(
        deck, "RSVD", renamer=RENAMERS["RSVD"], recordcountername="EQLNUM"
    )


def rvvd_fromdeck(
    deck: Union[str, "opm.libopmcommon_python.Deck"], ntequl: Optional[int] = None
) -> pd.DataFrame:
    """Extract RVVD data from a deck

    Args:
        deck
        ntequl: Number of EQLNUM regions in deck. Will
            be inferred if not present in deck
    """
    if "EQLDIMS" not in deck:
        deck = inferdims.inject_xxxdims_ntxxx("EQLDIMS", "NTEQUL", deck, ntequl)
    return common.ecl_keyworddata_to_df(
        deck, "RVVD", renamer=RENAMERS["RVVD"], recordcountername="EQLNUM"
    )


def pbvd_fromdeck(
    deck: Union[str, "opm.libopmcommon_python.Deck"], ntequl: Optional[int] = None
) -> pd.DataFrame:
    """Extract PBVD data from a deck

    Args:
        deck
        ntequl: Number of EQLNUM regions in deck. Will
            be inferred if not present in deck
    """
    if "EQLDIMS" not in deck:
        deck = inferdims.inject_xxxdims_ntxxx("EQLDIMS", "NTEQUL", deck, ntequl)
    return common.ecl_keyworddata_to_df(
        deck, "PBVD", renamer=RENAMERS["PBVD"], recordcountername="EQLNUM"
    )


def pdvd_fromdeck(
    deck: Union[str, "opm.libopmcommon_python.Deck"], ntequl: Optional[int] = None
) -> pd.DataFrame:
    """Extract PDVD data from a deck

    Args:
        deck
        ntequl: Number of EQLNUM regions in deck. Will
            be inferred if not present in deck
    """
    if "EQLDIMS" not in deck:
        deck = inferdims.inject_xxxdims_ntxxx("EQLDIMS", "NTEQUL", deck, ntequl)
    return common.ecl_keyworddata_to_df(
        deck, "PDVD", renamer=RENAMERS["PDVD"], recordcountername="EQLNUM"
    )


def phases_from_deck(deck: Union[str, "opm.libopmcommon_python.Deck"]) -> str:
    """Determined the set of phases from a deck, as
    a string with values "oil-water-gas", "gas-water", "oil-water",
    or "oil-gas"

    Args:
        deck: A parsed deck or DATA-file as a string

    Returns:
        String with phase configuration. Empty string if inconclusive.
    """
    if "OIL" in deck and "GAS" in deck and "WATER" in deck:
        return "oil-water-gas"
    if "OIL" not in deck and "GAS" in deck and "WATER" in deck:
        return "gas-water"
    if "OIL" in deck and "GAS" not in deck and "WATER" in deck:
        return "oil-water"
    if "OIL" in deck and "GAS" in deck and "WATER" not in deck:
        return "oil-gas"
    return ""


def phases_from_columns(columns: List[str]) -> str:
    """Determine the set of phases available in an
    equil dataframe, based on which columns are there.
    Returns "oil-water-gas", "gas-water", "oil-water",
    or "oil-gas"

    By this, we can pick the correct RENAMER to use.

    Args:
        columns: list of strings we can use for the determination

    Returns:
        String with phase configuration. Empty string if inconclusive.
    """
    if "OWC" in columns and "GOC" in columns:
        return "oil-water-gas"
    if "GWC" in columns and "OWC" not in columns and "GOC" not in columns:
        return "gas-water"
    if "OWC" in columns and "GOC" not in columns and "GWC" not in columns:
        return "oil-water"
    if "OWC" not in columns and "GOC" in columns and "GWC" not in columns:
        return "oil-gas"
    return ""


def equil_fromdeck(
    deck: Union[str, "opm.libopmcommon_python.Deck"], ntequl: Optional[int] = None
) -> pd.DataFrame:
    """Extract EQUIL data from a deck

    If the deck is supplied as a string object, the number
    of EQLNUM regions will be inferred if needed.

    Args:
        deck
        ntequl: Number of EQLNUM regions in deck.
    """
    if "EQLDIMS" not in deck:
        deck = inferdims.inject_xxxdims_ntxxx("EQLDIMS", "NTEQUL", deck, ntequl)

    phases = phases_from_deck(deck)
    if not phases or phases not in RENAMERS:
        raise ValueError(f"Could not determine phase configuration, got '{phases}'")
    columnrenamer = RENAMERS[phases_from_deck(deck)]

    dataframe = common.ecl_keyworddata_to_df(
        deck, "EQUIL", renamer=columnrenamer, recordcountername="EQLNUM"
    )

    # The column handling can be made prettier..
    for col in dataframe.columns:
        if "IGNORE" in col:
            del dataframe[col]

    return dataframe


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser (argparse.ArgumentParser or argparse.subparser): parser to
            fill with arguments
    """
    parser.add_argument("DATAFILE", help="Name of Eclipse DATA file.")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help=("Name of output csv file. " "Use '-' for stdout."),
        default="equil.csv",
    )
    parser.add_argument(
        "-k",
        "--keywords",
        nargs="+",
        help=(
            "List of EQUIL/SOLUTION keywords to include. "
            "If not supplied, all supported keywords will be included."
        ),
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def fill_reverse_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Fill a parser for the operation dataframe -> eclipse include file"""
    return common.fill_reverse_parser(parser, "EQUIL, RSVD++", "solution.inc")


def equil_main(args) -> None:
    """Read from disk and write CSV back to disk"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    if eclfiles:
        deck = eclfiles.get_ecldeck()
    if "EQLDIMS" in deck:
        # Things are easier when a full deck with (correct) EQLDIMS
        # is supplied:
        equil_df = df(deck, keywords=args.keywords)
    else:
        # This might be an include file for which we have to infer/guess
        # EQLDIMS. Then we send it to df() as a string
        equil_df = df(Path(args.DATAFILE).read_text())

    if "EQLNUM" in equil_df and "KEYWORD" in equil_df:
        eqlnums = str(len(equil_df["EQLNUM"].unique()))
        keywords = str(equil_df["KEYWORD"].unique())
    else:
        eqlnums = "-"
        keywords = "-"
    common.write_dframe_stdout_file(
        equil_df,
        args.output,
        index=False,
        caller_logger=logger,
        logstr=f"Unique EQLNUMs: {eqlnums}, keywords: {keywords}",
    )


def equil_reverse_main(args) -> None:
    """Entry-point for module, for command line utility for CSV to Eclipse"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    equil_df = pd.read_csv(args.csvfile)
    logger.info("Parsed %s", args.csvfile)
    inc_string = df2ecl(equil_df, keywords=args.keywords)
    common.write_inc_stdout_file(inc_string, args.output)


def df2ecl(
    equil_df: pd.DataFrame,
    keywords: Optional[List[str]] = None,
    comments: Optional[Dict[str, str]] = None,
    withphases: bool = False,
    filename: Optional[str] = None,
) -> str:
    """Generate Eclipse include strings from dataframes with
    solution (EQUIL, RSVD++) data.

    Args:
        equil_df: Dataframe with data on ecl2df format.
        keywords: List of keywords to include. Must be
            supported and present in the incoming dataframe.
        comments: Dictionary indexed by keyword with comments to be
            included pr. keyword. If a key named "master" is present
            it will be used as a master comment for the outputted file.
        withphases: If True, the phase configuration keywords
            will be outputted. This is mostly for testing, and should
            not be necessary in production.
        filename: If supplied, the generated text will also be dumped
            to file.

    """
    string = ""
    if withphases:
        string += (
            phases_from_columns(equil_df.columns).upper().replace("-", "\n") + "\n\n"
        )
    string += common.df2ecl(
        equil_df,
        keywords=keywords,
        comments=comments,
        supported=SUPPORTED_KEYWORDS,
        consecutive="EQLNUM",
        filename=filename,
    )
    return string


def df2ecl_equil(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print EQUIL keyword with data

    Args:
        dframe: Containing EQUIL data
        comment: Text that will be included as a comment
    """
    if dframe.empty:
        return "-- No data!"
    string = "EQUIL\n"
    string += common.comment_formatter(comment)

    if "KEYWORD" not in dframe:
        # Use everything..
        subset = dframe
    else:
        subset = dframe[dframe["KEYWORD"] == "EQUIL"]
    if "EQLNUM" not in subset:
        if len(subset) != 1:
            logger.critical("If EQLNUM is not supplied, only one row should be given")
            return ""
        subset["EQLNUM"] = 1
    subset = subset.set_index("EQLNUM").sort_index()

    phases = phases_from_columns(subset.columns)

    # Make a copy as we are going to modify it in order to have Pandas
    # make a pretty txt table:
    equildf = subset.copy()
    # Column names are pr. ec2ldf standard, redo to opm.common in order
    # to use sorting from that:
    inv_renamer = {value: key for key, value in RENAMERS[phases].items()}
    # print(inv_renamer)
    equildf.rename(inv_renamer, axis="columns", inplace=True)
    col_headers = [item["name"] for item in common.OPMKEYWORDS["EQUIL"]["items"]]
    for colname in col_headers:
        # Add those that are missing, as Eclipse defaults
        if colname not in equildf:
            equildf[colname] = "1*"
    # Reorder columns:
    equildf = equildf[col_headers]

    # It is critical for opm.common, maybe also E100 to have integers printed
    # as integers, for correct parsing. Ensure integer types where
    # the json says integer:
    integer_cols = [
        item["name"]
        for item in common.OPMKEYWORDS["EQUIL"]["items"]
        if item["value_type"] == "INT"
    ]
    for int_col in integer_cols:
        # But allow these columns to contain "1*"
        if set(equildf[int_col]) != {"1*"}:
            equildf[int_col] = equildf[int_col].astype(int)

    # Now rename again to have prettier column names:
    equildf.rename(RENAMERS[phases], axis="columns", inplace=True)
    # Add a final column with the end-slash, invisible header:
    equildf[" "] = "/"
    string += "-- " + equildf.to_string(header=True, index=False)
    return string + "\n\n"


def df2ecl_rsvd(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print RSVD keyword with data

    This data consists of one table (rs as a function
    of depth) for each EQLNUM

    Args:
        dframe: Containing RSVD data
        comment Text that will be included as a comment
    """
    return _df2ecl_equilfuncs("RSVD", dframe, comment)


def df2ecl_rvvd(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print RVVD keyword with data

    This data consists of one table (rv as a function
    of depth) for each EQLNUM

    Args:
        dframe: Containing RVVD data
        comment: Text that will be included as a comment
    """
    return _df2ecl_equilfuncs("RVVD", dframe, comment)


def df2ecl_pbvd(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print PBVD keyword with data

    Bubble-point versus depth

    This data consists of one table (Pb as a function
    of depth) for each EQLNUM

    Args:
        dframe: Containing PBVD data
        comment: Text that will be included as a comment
    """
    return _df2ecl_equilfuncs("PBVD", dframe, comment)


def df2ecl_pdvd(dframe: pd.DataFrame, comment: Optional[str] = None):
    """Print PDVD keyword with data.

    Dew-point versus depth.

    This data consists of one table (Pd as a function
    of depth) for each EQLNUM

    Args:
        dframe: Containing PDVD data
        comment: Text that will be included as a comment
    """
    return _df2ecl_equilfuncs("PDVD", dframe, comment)


def _df2ecl_equilfuncs(
    keyword: str, dframe: pd.DataFrame, comment: Optional[str] = None
) -> str:
    """Internal function to be used by df2ecl_<keyword>() functions"""
    if dframe.empty:
        return "-- No data!"
    string = f"{keyword}\n"
    string += common.comment_formatter(comment)
    col_headers = RENAMERS[keyword]["DATA"]

    string += f"--   {'DEPTH':^21} {col_headers[1]:^21} \n"
    if "KEYWORD" not in dframe:
        # Use everything..
        subset = dframe
    else:
        subset = dframe[dframe["KEYWORD"] == keyword]

    def _df2ecl_equilfuncs_eqlnum(dframe: pd.DataFrame) -> str:
        """Print one equilibriation function table for a specific
        EQLNUM

        Args:
            dframe (pd.DataFrame): Cropped to only contain data for one EQLNUM

        Returns:
            string
        """
        string = ""
        dframe = dframe.sort_values("Z")
        for _, row in dframe.iterrows():
            string += f"  {row[col_headers[0]]:20.7f} {row[col_headers[1]]:20.7f}\n"
        return string + "/\n"

    subset = subset.set_index("EQLNUM").sort_index()
    for eqlnum in subset.index.unique():
        string += f"-- EQLNUM: {eqlnum}\n"
        string += _df2ecl_equilfuncs_eqlnum(subset[subset.index == eqlnum])
    return string + "\n"



"""Extract GRUPTREE information from an Eclipse deck"""

import argparse
import collections
import datetime
import logging
import sys
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
import treelib

try:
    # Needed for mypy

    # pylint: disable=unused-import
    import opm.io
except ImportError:
    pass

from ecl2df import EclFiles, getLogger_ecl2csv
from ecl2df.common import (
    parse_opmio_date_rec,
    parse_opmio_deckrecord,
    parse_opmio_tstep_rec,
    write_dframe_stdout_file,
)

logger = logging.getLogger(__name__)


def df(
    deck: Union[EclFiles, "opm.libopmcommon_python.Deck"],
    startdate: Optional[datetime.date] = None,
    welspecs: bool = True,
) -> pd.DataFrame:
    """Extract all group information from a deck
    and present as a Pandas Dataframe of all edges.

    Properties for nodes given in GRUPNET/NODEPROP will
    be added as extra columns.

    From WELSPECS, well names are extracted and added
    as nodes with an edge to its group.

    The gruptree is a time dependent property,
    with accumulative effects from new occurences of
    GRUPNET, WELSPECS, BRANPROP and NODEPROP.

    Whenever the GRUPTREE or BRANPROP tree changes, the
    previous tree is copied and a new complete tree is added
    to the dataframe tagged with the new date.

    startdate is only relevant when START is not in the deck.

    Args:
        deck: opm.io Deck object or EclFiles

    Returns:
        pd.DataFrame with one row pr edge. Empty dataframe if no
        information is found in deck.
    """

    date: Optional[datetime.date]
    if startdate is not None:
        date = startdate
    else:
        date = None

    if isinstance(deck, EclFiles):
        deck = deck.get_ecldeck()

    edgerecords = []  # list of dict of rows containing an edge.
    nodedatarecords = []

    # In order for the GRUPTREE/BRANPROP keywords to accumulate, we
    # store the edges as dictionaries indexed by the edge
    # (which is a tuple of child and parent).
    currentedges: Dict[str, Dict[Tuple[str, str], Dict[str, Any]]] = {
        "GRUPTREE": {},
        "BRANPROP": {},
    }
    # Same approach for the welspecs keywords
    wellspecsedges: Dict[Tuple[str, str], str] = {}
    # Node properties from GRUPNET/NODEPROP is stored in a dataframe
    # Note that it's not allowed to mix GRUPNET and NODEPROP in eclipse
    # so the datframe will only contain columns from one of them
    nodedata: Dict[str, pd.DataFrame] = {
        "GRUPNET": pd.DataFrame(),
        "NODEPROP": pd.DataFrame(),
    }

    # Flags which will tell when a new network related keyword
    # has been encountered
    keywords = ["GRUPTREE", "BRANPROP", "WELSPECS", "GRUPNET", "NODEPROP"]
    found_keywords = {key: False for key in keywords}
    for kword in deck:
        if kword.name in ["DATES", "START", "TSTEP"]:
            # Whenever we encounter a new DATES, it means that
            # we have processed all the network keywords that
            # have occured since the last date, so this is the chance
            # to dump the parsed data. Also we dump the *entire* tree
            # at every date with a change, not only the newfound edges.
            if any(val for val in found_keywords.values()):
                if date is None:
                    logger.warning("No date parsed, maybe you should pass --startdate")
                    logger.warning("Using 1900-01-01")
                    date = datetime.date(year=1900, month=1, day=1)
                edgerecords += _write_edgerecords(
                    currentedges, nodedata, wellspecsedges, found_keywords, date
                )
                found_keywords = {key: False for key in keywords}
            # Done dumping the data for the previous date, parse the fresh
            # date:
            if kword.name in ["DATES", "START"]:
                for rec in kword:
                    date = parse_opmio_date_rec(rec)
                    logger.debug("Parsing at date %s", str(date))
            elif kword.name == "TSTEP":
                assert date is not None
                for rec in kword:
                    steplist = parse_opmio_tstep_rec(rec)
                    # Assuming not LAB units, then the unit is days.
                    days = sum(steplist)
                    date += datetime.timedelta(days=days)
                    logger.info(
                        "Advancing %s days to %s through TSTEP", str(days), str(date)
                    )
        if kword.name in ["GRUPTREE", "BRANPROP"]:
            found_keywords[kword.name] = True
            renamer = (
                {"DOWNTREE_NODE": "CHILD_GROUP", "UPTREE_NODE": "PARENT_GROUP"}
                if kword.name == "BRANPROP"
                else None
            )
            for edgerec in kword:
                edge_dict = parse_opmio_deckrecord(edgerec, kword.name, renamer=renamer)
                child = edge_dict.pop("CHILD_GROUP")
                parent = edge_dict.pop("PARENT_GROUP")
                currentedges[kword.name][(child, parent)] = edge_dict

        if kword.name == "WELSPECS" and welspecs:
            found_keywords["WELSPECS"] = True
            for wellrec in kword:
                wspc_dict = parse_opmio_deckrecord(wellrec, "WELSPECS")
                wellspecsedges[(wspc_dict["WELL"], wspc_dict["GROUP"])] = "WELSPECS"

        if kword.name in ["GRUPNET", "NODEPROP"]:
            found_keywords[kword.name] = True
            renamer = (
                {"PRESSURE": "TERMINAL_PRESSURE"} if kword.name == "NODEPROP" else None
            )
            for rec in kword:
                nodedatarecords.append(
                    parse_opmio_deckrecord(rec, kword.name, renamer=renamer)
                )
            nodedata[kword.name] = (
                pd.DataFrame(nodedatarecords)
                .drop_duplicates(subset="NAME", keep="last")
                .set_index("NAME")
            )

    # Ensure we also store any tree information found after the last DATE statement
    if any(val for val in found_keywords.values()):
        edgerecords += _write_edgerecords(
            currentedges, nodedata, wellspecsedges, found_keywords, date
        )
    dframe = pd.DataFrame(edgerecords)
    if "DATE" in dframe:
        dframe["DATE"] = pd.to_datetime(dframe["DATE"])

    # Remove rows with duplicate DATE, CHILD and KEYWORD
    # This happens with WELSPECS if both GRUPTREE and BRANPROP is defined
    # at the same timestep. And when a node is redirected to a new parent node
    dframe = dframe.drop_duplicates(subset=["DATE", "CHILD", "KEYWORD"], keep="last")
    print(dframe)
    return dframe


def _write_edgerecords(
    currentedges: Dict[str, Dict[Tuple[str, str], Dict[str, Any]]],
    nodedata: Dict[str, pd.DataFrame],
    wellspecsedges: Dict[Tuple[str, str], str],
    found_keywords: dict,
    date: Optional[datetime.date],
) -> List[dict]:
    """Writes a new GRUPTREE tree if there are new instances of
    GRUPTREE, GRUPNET or WELSPECS and writes a new BRANPROP tree
    if there are new instances of BRANPROP, NODEPROP or WELSPECS.

    If both trees are written, any duplicate WELSPECS rows are
    removed at the end of the df function.
    """
    edgerecords = []
    for treetype, nodetype, welspecs in [
        ("GRUPTREE", "GRUPNET", "WELSPECS"),
        ("BRANPROP", "NODEPROP", "WELSPECS"),
    ]:
        if any(found_keywords[key] for key in [treetype, nodetype, welspecs]):
            edgerecords += _merge_edges_and_nodeinfo(
                currentedges[treetype],
                nodedata[nodetype],
                wellspecsedges,
                date,
                treetype,
            )
    return edgerecords


def _merge_edges_and_nodeinfo(
    currentedges: Dict[Tuple[str, str], Dict[str, Any]],
    nodedata_df: pd.DataFrame,
    wellspecsedges: Dict[Tuple[str, str], str],
    date: Optional[datetime.date],
    treetype: str,
) -> List[dict]:
    """Merge a list of edges with information from the nodedata dataframe.

    Edges where there is no parent (root nodes) are identified and added
    as special cases.

    Welspecs-edges are allowed to have a parent that is missing from the
    GRUPTREE tree, but these edges are ignored for BRANPROP trees.

    Args:
        currentedges:
        nodedata_df: Containing data for each node to add.
        date: Relevant date.

    Returns:
        List of dictionaries (that can be made into a dataframe)
    """
    edgerecords = []
    childs = set()
    parents = set()
    # Write GRUPTREE/BRANPROP edges
    for (child, parent), edge_dict in currentedges.items():
        rec_dict = {"DATE": date, "CHILD": child, "PARENT": parent, "KEYWORD": treetype}
        childs |= {child}
        parents |= {parent}
        # Add fields from edge_dict
        rec_dict.update(edge_dict)
        # Add node data
        if child in nodedata_df.index:
            rec_dict.update(nodedata_df.loc[child])
        edgerecords.append(rec_dict)

    # Write WELSPECS edges
    welspecs_parents = set()
    for (child, parent), _ in wellspecsedges.items():
        # For BRANPROP trees, only wells with a parent in the tree are added
        if (treetype == "BRANPROP" and parent in childs) or (treetype == "GRUPTREE"):
            rec_dict = {
                "DATE": date,
                "CHILD": child,
                "PARENT": parent,
                "KEYWORD": "WELSPECS",
            }
            edgerecords.append(rec_dict)
        if treetype == "GRUPTREE":
            welspecs_parents |= {parent}

    # Missing welspecs parents are connected directly to the FIELD node
    # (only for GRUPTREE)
    for parent in welspecs_parents - childs:
        edgerecords.append(
            {"DATE": date, "CHILD": parent, "KEYWORD": treetype, "PARENT": "FIELD"}
        )
        parents |= {"FIELD"}

    roots = parents - childs
    rootrecords = []
    for root in roots:
        rec_dict = {"DATE": date, "CHILD": root, "KEYWORD": treetype}
        if root in nodedata_df.index:
            rec_dict.update(nodedata_df.loc[root])
        rootrecords.append(rec_dict)
    return rootrecords + edgerecords


def edge_dataframe2dict(dframe: pd.DataFrame) -> List[dict]:
    """Convert list of edges in a dataframe into a
    nested dictionary (tree).

    The dataframe cannot have multiple DATEs, filter to the date you
    want prior to calling this function.

    Example::

      [{
        'FIELD': {'OP': {'OP_1': {},
        'OP_2': {},
        'OP_3': {},
        'OP_4': {},
        'OP_5': {}},
        'WI': {'WI_1': {}, 'WI_2': {}, 'WI_3': {}}}
      }]

    Leaf nodes have empty dictionaries as their value.

    Returns:
        List of nested dictionaries, as we in general
        may have more than one root. The list sorted
        alphabetically on the top level node name.
    """
    if dframe.empty:
        return [{}]
    if "DATE" in dframe:
        if len(dframe["DATE"].unique()) > 1:
            raise ValueError("Can only handle one date at a time")
    subtrees: dict = collections.defaultdict(dict)
    edges = []  # List of tuples
    for _, row in dframe.iterrows():
        if not pd.isnull(row["PARENT"]):
            edges.append((row["CHILD"], row["PARENT"]))
    for child, parent in edges:
        subtrees[parent][child] = subtrees[child]

    children, parents = zip(*edges)
    roots = set(parents).difference(children)
    return [{root: subtrees[root]} for root in sorted(roots)]


def _add_to_tree_from_dict(
    nested_dict: dict, name: str, tree: treelib.Tree, parent: Optional[str] = None
) -> None:
    assert isinstance(nested_dict, dict)
    tree.create_node(name, name, parent=parent)
    for key, value in sorted(nested_dict.items()):
        _add_to_tree_from_dict(nested_dict=value, name=key, tree=tree, parent=name)


def tree_from_dict(nested_dict: dict) -> treelib.Tree:
    """Convert a dictionary to a treelib Tree.

    The treelib representation of the trees is used
    for pretty-printing (in ASCII) of the tree, you
    only need to do str() of the returned result
    from this function.

    See `https://treelib.readthedocs.io/`

    Args:
        nested_dict: nested dictonary representing a tree.
    """
    if not nested_dict.keys():
        # Return an empty string, because str(treelib.Tree()) will error
        return ""
    if not len(nested_dict.keys()) == 1:
        raise ValueError(
            "The dict given to tree_from_dict() must have "
            "exactly one top level key, representing a single tree."
        )
    root_name = list(nested_dict.keys())[0]
    tree = treelib.Tree()
    _add_to_tree_from_dict(nested_dict[root_name], root_name, tree)
    return tree


def dict2treelib(name: str, nested_dict: dict) -> treelib.Tree:
    """Convert a nested dictonary to a treelib Tree
    object. This function is recursive.

    The treelib representation of the trees is used
    for pretty-printing (in ASCII) of the tree, you
    only need to do str() of the returned result
    from this function.

    See `https://treelib.readthedocs.io/`

    Args:
        name: name of root node
        nested_dict: nested dictonary of the children at the root.
    """
    warnings.warn(
        "dict2treelib() is deprecated and will be removed, use tree_from_dict()",
        FutureWarning,
    )

    tree = treelib.Tree()
    tree.create_node(name, name)
    for child in nested_dict.keys():
        tree.paste(name, dict2treelib(child, nested_dict[child]))
    return tree


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser (argparse.ArgumentParser or argparse.subparser): parser
            to fill with arguments
    """
    parser.add_argument("DATAFILE", help="Name of Eclipse DATA file.")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file. No CSV dump if empty",
        default="",
    )
    parser.add_argument(
        "-p",
        "--prettyprint",
        #      type=bool,
        action="store_true",
        help="Pretty-print the tree structure",
    )
    parser.add_argument(
        "--startdate",
        type=str,
        help="First schedule date if not defined in input file, YYYY-MM-DD",
        default=None,
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def prettyprint(dframe: pd.DataFrame) -> str:
    """Returns a string with the prettyprint of a dataframe
    possibly containing multiple dates and both GRUPTREE and
    BRANPROP trees"""
    output = ""
    for date in dframe["DATE"].dropna().unique():
        df_date = dframe[dframe.DATE == date]
        output += "Date: " + str(date.astype("M8[D]")) + "\n"

        for treetype in ["GRUPTREE", "BRANPROP"]:
            if treetype in df_date["KEYWORD"].unique():
                df_treetype = df_date[df_date["KEYWORD"].isin([treetype, "WELSPECS"])]
                if treetype == "BRANPROP":
                    # filter out edges where the parent is not a
                    # child in the tree. This will be printed as
                    # a separate tree for GRUPTREE
                    df_treetype = df_treetype[
                        df_treetype.PARENT.isin(df_treetype.CHILD.unique())
                    ]

                output += f"{treetype} trees:\n"
                for tree in edge_dataframe2dict(df_treetype):
                    output += str(tree_from_dict(tree))
                    output += "\n"
        output += "\n"
    return output


def gruptree_main(args) -> None:
    """Entry-point for module, for command line utility."""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    if not args.output and not args.prettyprint:
        print("Nothing to do. Set --output or --prettyprint")
        sys.exit(0)
    eclfiles = EclFiles(args.DATAFILE)
    dframe = df(eclfiles.get_ecldeck(), startdate=args.startdate)
    if args.prettyprint:
        if "DATE" in dframe:
            print(prettyprint(dframe))
        else:
            logger.warning("No tree data to prettyprint")
    elif args.output:
        write_dframe_stdout_file(dframe, args.output, index=False, caller_logger=logger)



"""Test module for equil2df"""

import os
import subprocess
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ecl2df import csv2ecl, ecl2csv, equil
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, nothing relevant in here then",
        allow_module_level=True,
    )


TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_equil2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(REEK)
    equildf = equil.df(eclfiles)
    expected = {}
    expected["EQUIL"] = pd.DataFrame(
        [
            {
                "Z": 2469.0,
                "PRESSURE": 382.4,
                "OWC": 1700.0,
                "PCOWC": 0.0,
                "GOC": 0.0,
                "PCGOC": 0.0,
                "INITRS": 1.0,
                "INITRV": 0.0,
                "OIP_INIT": 20.0,
                "EQLNUM": 1,
                "KEYWORD": "EQUIL",
            },
            {
                "Z": 2469.0,
                "PRESSURE": 382.4,
                "OWC": 1000.0,
                "PCOWC": 0.0,
                "GOC": 0.0,
                "PCGOC": 0.0,
                "INITRS": 2.0,
                "INITRV": 0.0,
                "OIP_INIT": 20.0,
                "EQLNUM": 2,
                "KEYWORD": "EQUIL",
            },
        ]
    )
    expected["RSVD"] = pd.DataFrame(
        [
            {"Z": 1500.0, "EQLNUM": 1, "KEYWORD": "RSVD", "RS": 184.0},
            {"Z": 4000.0, "EQLNUM": 1, "KEYWORD": "RSVD", "RS": 184.0},
            {"Z": 1500.0, "EQLNUM": 2, "KEYWORD": "RSVD", "RS": 184.0},
            {"Z": 4000.0, "EQLNUM": 2, "KEYWORD": "RSVD", "RS": 184.0},
        ]
    )

    for keyword, df in equildf.groupby("KEYWORD"):
        pd.testing.assert_frame_equal(
            df.dropna(axis=1).reset_index(drop=True), expected[keyword]
        )

    # Check that we can dump from dataframe to include file
    # and reparse to the same dataframe:
    inc = equil.df2ecl(equildf, withphases=True)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(equildf, df_from_inc, check_dtype=False)


def test_df2ecl(tmp_path):
    """Test that we can write include files to disk"""
    os.chdir(tmp_path)
    eclfiles = EclFiles(EIGHTCELLS)
    equildf = equil.df(eclfiles)
    equil.df2ecl(equildf, filename="equil.inc")
    assert Path("equil.inc").is_file()

    # Test automatic directory creation:
    equil.df2ecl(equildf, filename="eclipse/include/equil.inc")
    assert Path("eclipse/include/equil.inc").is_file()


def test_df2ecl_equil(tmp_path):
    """Test the underlying function directly"""
    dframe = pd.DataFrame(
        [
            {
                "Z": 2469.0,
                "PRESSURE": 382.4,
                "OWC": 1700.0,
                "PCOWC": 0.0,
                "GOC": 0.0,
                "EQLNUM": 1,
                "KEYWORD": "EQUIL",
            },
        ]
    )
    # Check that we don't need the KEYWORD in the underlying function
    assert equil.df2ecl_equil(dframe) == equil.df2ecl_equil(
        dframe.drop("KEYWORD", axis="columns")
    )

    # Can also drop EQLNUM since we have only one row:
    assert equil.df2ecl_equil(dframe) == equil.df2ecl_equil(
        dframe.drop("EQLNUM", axis="columns")
    )

    # Problem if we have two rows, nothing is returned and a critical error is logged
    assert (
        equil.df2ecl_equil(pd.concat([dframe, dframe]).drop("EQLNUM", axis="columns"))
        == ""
    )


def test_decks():
    """Test some string decks"""
    deckstr = """
OIL
WATER
GAS

EQUIL
 2000 200 2200 /
"""
    df = equil.df(deckstr)
    assert df["OWC"].values == 2200
    assert len(df) == 1
    assert "IGNORE1" not in df
    assert df["EQLNUM"].unique()[0] == 1
    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    # 0 columns can be both integers and floats.
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)

    # Test empty data:
    inc = equil.df2ecl_equil(equil.df(""))
    assert "No data" in inc
    assert equil.df(inc).empty

    # Test more empty data:
    assert "No data" in equil.df2ecl_equil(equil.df(""))
    assert "No data" in equil.df2ecl_rsvd(equil.df(""))
    assert "No data" in equil.df2ecl_rvvd(equil.df(""))
    assert "No data" in equil.df2ecl_pbvd(equil.df(""))
    assert "No data" in equil.df2ecl_pdvd(equil.df(""))

    deckstr = """
OIL
WATER

EQUIL
 2000 200 2200 /
"""
    df = equil.df(deckstr)
    assert df["OWC"].values == 2200
    assert len(df) == 1
    assert "IGNORE1" not in df
    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    # 0 columns can be both integers and floats.
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)

    deckstr = """
GAS
WATER

EQUIL
 2000 200 2200 /
"""
    df = equil.df(deckstr)
    assert df["GWC"].values == 2200
    assert "OWC" not in df
    assert len(df) == 1
    assert "IGNORE2" not in df
    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    # 0 columns can be both integers and floats.
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)

    deckstr = """
GAS
OIL

EQUIL
 2000 200 2200 1 2100 3 /
"""
    df = equil.df(deckstr)
    assert df["GOC"].values == 2100
    assert "GWC" not in df
    assert "OWC" not in df
    assert len(df) == 1
    assert "IGNORE2" not in df
    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    # 0 columns can be both integers and floats.
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)

    deckstr = """
OIL
WATER
GAS

-- Output file printed by ecl2df.equil 0.5.2.dev12+g785dc0d.d20200402
-- at 2020-04-03 16:18:57.450100

EQUIL
--   DATUM  PRESSURE     OWC  PCOWC  GOC  PCGOC  INITRS  INITRV  ACCURACY
 2469.0     382.4  1700.0    0.0  0.0    0.0     1     0      20  /
 2469.0     382.4  1000.0    0.0  0.0    0.0     2     0      20  /
"""
    df = equil.df(deckstr)
    assert set(df["GOC"].values) == {0.0}
    assert "GWC" not in df
    assert "OWC" in df
    assert len(df) == 2
    assert "IGNORE2" not in df
    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    # 0 columns can be both integers and floats.
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)


def test_equil_fromdeck():
    """equil.df relies on equil.equil_fromdeck, test that expliclitly"""
    deckstr = """
OIL
WATER
GAS

EQUIL
--   DATUM  PRESSURE     OWC  PCOWC  GOC  PCGOC  INITRS  INITRV  ACCURACY
 2469.0     382.4  1700.0    0.0  0.0    0.0     1     0      20  /
 2469.0     382.4  1000.0    0.0  0.0    0.0     2     0      20  /
"""
    pd.testing.assert_frame_equal(
        equil.equil_fromdeck(deckstr), equil.df(deckstr).drop("KEYWORD", axis=1)
    )
    # If we supply a deck object and not a string, it will not be able
    # to pick up both EQLNUMs:
    assert len(equil.equil_fromdeck(deckstr)) == 2  # correct
    assert len(equil.equil_fromdeck(deckstr, 2)) == 2
    assert len(equil.equil_fromdeck(deckstr, 1)) == 1
    assert len(equil.equil_fromdeck(EclFiles.str2deck(deckstr))) == 1  # (watch out!)

    wrongdeck = """
EQUIL
1 1 1 1 1 1 1 1 1 /
/
"""
    with pytest.raises(ValueError, match="Could not determine phase configuration"):
        equil.equil_fromdeck(wrongdeck)

    with pytest.raises(ValueError, match="Could not determine phase configuration"):
        equil.equil_fromdeck("")

    # Single phase decks will not work:
    gasdeck = """
GAS

EQUIL
1 1 5000 0 4000 /
"""
    with pytest.raises(ValueError, match="Could not determine phase configuration"):
        equil.equil_fromdeck(gasdeck)


def test_rsvd():
    """Test RSVD tables"""
    deckstr = """
RSVD
 10 100 /
 30 400 /
 50 100 /"""
    rsvd_df = equil.df(deckstr)
    assert "KEYWORD" in rsvd_df
    assert "EQUIL" not in rsvd_df["KEYWORD"].values
    assert max(rsvd_df["EQLNUM"]) == 3
    assert set(rsvd_df["Z"].values) == {10, 30, 50}
    assert set(rsvd_df["RS"].values) == {100, 400}
    inc = equil.df2ecl(rsvd_df)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(rsvd_df, df_from_inc)

    assert equil.df(deckstr, keywords="EQUIL").empty

    # Check that we can use the underlying function directly:
    rsvd_df2 = equil.rsvd_fromdeck(deckstr)
    pd.testing.assert_frame_equal(rsvd_df.drop("KEYWORD", axis="columns"), rsvd_df2)

    deckstr = """
RSVD
 10 100
 30 400 /
 50 100
 60 1000 /"""
    rsvd_df = equil.df(deckstr)
    assert "KEYWORD" in rsvd_df
    assert "EQUIL" not in rsvd_df["KEYWORD"].values
    assert len(rsvd_df) == 4
    assert max(rsvd_df["EQLNUM"]) == 2
    assert set(rsvd_df["Z"].values) == {10, 30, 50, 60}
    assert set(rsvd_df["RS"].values) == {100, 400, 1000}
    inc = equil.df2ecl(rsvd_df)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(rsvd_df, df_from_inc)

    # Check that we can use the underlying function directly:
    rsvd_df2 = equil.rsvd_fromdeck(deckstr)
    pd.testing.assert_frame_equal(rsvd_df.drop("KEYWORD", axis="columns"), rsvd_df2)


def test_rvvd():
    """Test RVVD tables"""
    deckstr = """
RVVD
 10 100 /
 30 400 /
 50 100 /"""
    rvvd_df = equil.df(deckstr)
    assert "KEYWORD" in rvvd_df
    assert "EQUIL" not in rvvd_df["KEYWORD"].values
    assert max(rvvd_df["EQLNUM"]) == 3
    assert set(rvvd_df["Z"].values) == {10, 30, 50}
    assert set(rvvd_df["RV"].values) == {100, 400}

    inc = equil.df2ecl(rvvd_df)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(rvvd_df, df_from_inc)

    assert equil.df(deckstr, keywords="EQUIL").empty

    # Check that we can use the underlying function directly:
    rvvd_df2 = equil.rvvd_fromdeck(deckstr)
    pd.testing.assert_frame_equal(rvvd_df.drop("KEYWORD", axis="columns"), rvvd_df2)

    deckstr = """
RVVD
 10 100
 30 400 /
 50 100
 60 1000 /"""
    rvvd_df = equil.df(deckstr)
    assert "KEYWORD" in rvvd_df
    assert "EQUIL" not in rvvd_df["KEYWORD"].values
    assert len(rvvd_df) == 4
    assert max(rvvd_df["EQLNUM"]) == 2
    assert set(rvvd_df["Z"].values) == {10, 30, 50, 60}
    assert set(rvvd_df["RV"].values) == {100, 400, 1000}

    inc = equil.df2ecl(rvvd_df)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(rvvd_df, df_from_inc)


def test_pbvd():
    """Test PBVD tables"""
    deckstr = """
PBVD
 10 100 /
 30 400 /
 50 100 /"""
    pbvd_df = equil.df(deckstr)
    assert "KEYWORD" in pbvd_df
    assert "EQUIL" not in pbvd_df["KEYWORD"].values
    assert max(pbvd_df["EQLNUM"]) == 3
    assert set(pbvd_df["Z"].values) == {10, 30, 50}
    assert set(pbvd_df["PB"].values) == {100, 400}

    inc = equil.df2ecl(pbvd_df)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(pbvd_df, df_from_inc)

    assert equil.df(deckstr, keywords="EQUIL").empty

    # Check that we can use the underlying function directly:
    pbvd_df2 = equil.pbvd_fromdeck(deckstr)
    pd.testing.assert_frame_equal(pbvd_df.drop("KEYWORD", axis="columns"), pbvd_df2)

    # Check that we don't need the KEYWORD column for the underlying function:
    assert equil.df2ecl_pbvd(pbvd_df) == equil.df2ecl_pbvd(
        pbvd_df.drop("KEYWORD", axis="columns")
    )

    # If EQLNUM column is dropped it is not possible to guess the
    # correct include file, so the code must fail:
    with pytest.raises(KeyError):
        equil.df2ecl_pbvd(pbvd_df.drop("EQLNUM", axis="columns"))


def test_pdvd():
    """Test PDVD tables"""
    deckstr = """
PDVD
 10 100 /
 30 400 /
 50 100 /"""
    pdvd_df = equil.df(deckstr)
    assert "KEYWORD" in pdvd_df
    assert "EQUIL" not in pdvd_df["KEYWORD"].values
    assert max(pdvd_df["EQLNUM"]) == 3
    assert set(pdvd_df["Z"].values) == {10, 30, 50}
    assert set(pdvd_df["PD"].values) == {100, 400}

    inc = equil.df2ecl(pdvd_df)
    df_from_inc = equil.df(inc)
    pdvd_df2 = equil.pdvd_fromdeck(deckstr)
    pd.testing.assert_frame_equal(pdvd_df, df_from_inc)

    assert equil.df(deckstr, keywords="EQUIL").empty

    # Check that we can use the underlying function directly:
    pdvd_df2 = equil.pdvd_fromdeck(deckstr)
    pd.testing.assert_frame_equal(pdvd_df.drop("KEYWORD", axis="columns"), pdvd_df2)


def test_rsvd_via_file(tmp_path, mocker):
    """Test that we can reparse RSVD with unknown TABDIMS
    from a file using the command line utility"""
    os.chdir(tmp_path)
    deckstr = """
RSVD
 10 100
 30 400 /
 50 100
 60 1000 /"""
    rsvd_df = equil.df(deckstr)
    with open("rsvd.inc", "w") as filehandle:
        filehandle.write(deckstr)
    mocker.patch("sys.argv", ["ecl2csv", "equil", "-v", "rsvd.inc", "-o", "rsvd.csv"])
    ecl2csv.main()
    rsvd_df_fromcsv = pd.read_csv("rsvd.csv")
    pd.testing.assert_frame_equal(rsvd_df, rsvd_df_fromcsv)


def test_ntequl():
    """Test that we can infer NTEQUL when not supplied"""
    deckstr = """
GAS
OIL

EQUIL
 2000 200 2200 1 2100 3 /
 3000 200 2200 1 2100 3 /
"""
    df = equil.df(deckstr)
    assert set(df["GOC"].values) == set([2100, 2100])
    assert len(df) == 2
    assert df["EQLNUM"].min() == 1
    assert df["EQLNUM"].max() == 2
    # Supply correct NTEQUL instead of estimating
    df = equil.df(deckstr, ntequl=2)
    assert len(df) == 2

    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)

    # Supplying wrong NTEQUIL:
    df = equil.df(deckstr, ntequl=1)
    # We are not able to catch this situation..
    assert len(df) == 1
    # But this will fail:
    with pytest.raises(ValueError):
        equil.df(deckstr, ntequl=3)

    deckstr = """
GAS
OIL

EQLDIMS
 2 /

EQUIL
 2000 200 2200 1 2100 3 /
 3000 200 2200 1 2100 3 /
"""
    df = equil.df(deckstr)
    assert set(df["GOC"].values) == set([2100, 2100])
    assert len(df) == 2

    inc = equil.df2ecl(df, withphases=True)
    df_from_inc = equil.df(inc)
    pd.testing.assert_frame_equal(df, df_from_inc, check_dtype=False)


@pytest.mark.parametrize(
    "somefloat, expected",
    [
        (1000.00000000000000000005, " 1000.0 "),
        (1000.0000000000003, " 1000.0 "),  # Many decimals may destabilize Eclipse
        (1000.0000003, " 1000.0 "),
        (1000.000003, " 1000.000003 "),  # Assume Eclipse accepts this
        (np.float32(1000.00003), " 1000.0 "),
        (np.float32(1000.0003), " 1000.0003"),  # can give 1000.000305
    ],
)
def test_eclipse_rounding(somefloat, expected):
    """Values in include files with a lot of decimals, like you sometimes get
    from Python floating point operations may crash Eclipse. Ensure these are
    rounded (the typical output routine is pd.DataFrame.to_string())

    As an example, 1000.00000000000000000005 as a datum value in EQUIL has been
    observed to crash Eclipse. It would be hard to get Python to output this
    particular value.
    """

    dframe = pd.DataFrame(
        [
            {
                "Z": 2469.0,
                "PRESSURE": 382.4,
                "OWC": somefloat,
                "PCOWC": 0.0,
                "GOC": 0.0,
                "PCGOC": 0.0,
                "INITRS": 1.0,
                "INITRV": 0.0,
                "EQLNUM": 1,
                "KEYWORD": "EQUIL",
            }
        ]
    )
    assert expected in equil.df2ecl(dframe, withphases=False)


def test_main_subparser(tmp_path, mocker, capsys):
    """Test command line interface"""
    os.chdir(tmp_path)
    tmpcsvfile = "equil.csv"
    mocker.patch("sys.argv", ["ecl2csv", "equil", "-v", REEK, "-o", tmpcsvfile])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(tmpcsvfile)
    assert not disk_df.empty

    # Test the reverse operation:
    mocker.patch(
        "sys.argv", ["csv2ecl", "equil", "-v", "--output", "equil.inc", tmpcsvfile]
    )
    csv2ecl.main()
    # NB: cvs2ecl does not output the phase configuration!
    phases = "WATER\nGAS\nOIL\n\n"
    ph_equil_inc = Path("phasesequil.inc")
    ph_equil_inc.write_text(phases + Path("equil.inc").read_text())

    pd.testing.assert_frame_equal(equil.df(ph_equil_inc.read_text()), disk_df)

    # Test via stdout:
    result = subprocess.run(
        ["csv2ecl", "equil", "--output", "-", tmpcsvfile],
        stdout=subprocess.PIPE,
    )
    pd.testing.assert_frame_equal(
        equil.df(phases + result.stdout.decode()),
        disk_df,
        check_like=True,
    )

    # Test empty equil data:
    Path("poro.inc").write_text(
        """
GAS
OIL

PORO
0.1 0.1 /
"""
    )
    mocker.patch("sys.argv", ["ecl2csv", "equil", "-v", "poro.inc", "-o", "empty.csv"])
    ecl2csv.main()
    assert not Path("empty.csv").read_text().strip()


@pytest.mark.parametrize(
    "deckstring, expected",
    [
        ("", ""),
        ("OIL", ""),
        ("OIL\nWATER", "oil-water"),
        ("WATER\nOIL", "oil-water"),
        ("WATER", ""),
        ("GAS", ""),
        ("WATER\nGAS", "gas-water"),
        ("GAS\nWATER", "gas-water"),
        ("OIL\nWATER\nGAS", "oil-water-gas"),
        ("oil\nwater\ngas", "oil-water-gas"),  # (!) OPM is case insensitive
        ("OIL\nWATER\ngas", "oil-water-gas"),
    ],
)
def test_phases_from_deck(deckstring, expected):
    """Test that we can extract phase configuration from a deck"""
    deck = EclFiles.str2deck(deckstring)
    assert equil.phases_from_deck(deck) == expected


@pytest.mark.parametrize(
    "stringlist, expected",
    [
        ([""], ""),
        (["OWC"], "oil-water"),
        (["GWC"], "gas-water"),
        (["GOC"], "oil-gas"),
        (["OWC", "GOC"], "oil-water-gas"),
        (["OWC", "GOC"], "oil-water-gas"),
        (["OWC", "GWC"], ""),  # Meaningless
        (["GOC", "GWC"], ""),
        (["WOC"], ""),  # common typo..
    ],
)
def test_phases_from_columns(stringlist, expected):
    """Test that we can pick the correct phase configuration based
    on the column names in a dataframe."""
    assert equil.phases_from_columns(stringlist) == expected



"""
Extract saturation function data (SWOF, SGOF, SWFN, etc.)
from an Eclipse deck as Pandas DataFrame.

Data can be extracted from a full Eclipse deck (`*.DATA`)
or from individual files.

Note that when parsing from individual files, it is
undefined in the syntax how many saturation functions (SATNUMs) are
present. For convenience, it is possible to estimate the count of
SATNUMs, but whenever this is known, it is recommended to either supply
TABDIMS or to supply the satnumcount directly to avoid possible bugs.

"""
import argparse
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

import pandas as pd

try:
    # pylint: disable=unused-import
    import opm.io
except ImportError:
    pass

from ecl2df import common, getLogger_ecl2csv, inferdims

from .common import write_dframe_stdout_file
from .eclfiles import EclFiles

logger: logging.Logger = logging.getLogger(__name__)

SUPPORTED_KEYWORDS: List[str] = [
    "SWOF",
    "SGOF",
    "SWFN",
    "SGWFN",
    "SOF2",
    "SGFN",
    "SOF3",
    "SLGOF",
]

# RENAMERS are a dictionary of dictionaries, referring to
# how we should rename deck record items, from the JSON
# files in opm.common and into Dataframe column names.
RENAMERS: Dict[str, Dict[str, Union[List[str], str]]] = {}
RENAMERS["SGFN"] = {"DATA": ["SG", "KRG", "PCOG"]}
RENAMERS["SGOF"] = {"DATA": ["SG", "KRG", "KROG", "PCOG"]}
RENAMERS["SGWFN"] = {"DATA": ["SG", "KRG", "KRW", "PCGW"]}
RENAMERS["SLGOF"] = {"DATA": ["SL", "KRG", "KRO", "PCOG"]}
RENAMERS["SOF2"] = {"DATA": ["SO", "KRO"]}
RENAMERS["SOF3"] = {"DATA": ["SO", "KROW", "KROG"]}
RENAMERS["SWFN"] = {"DATA": ["SW", "KRW", "PCOW"]}
RENAMERS["SWOF"] = {"DATA": ["SW", "KRW", "KROW", "PCOW"]}


def df(
    deck: Union[str, "opm.libopmcommon_python.Deck"],
    keywords: Optional[List[str]] = None,
    ntsfun: Optional[int] = None,
) -> pd.DataFrame:
    """Extract the data in the saturation function keywords as a Pandas
    DataFrames.

    Data for all saturation functions are merged into one dataframe.
    The two first columns in the dataframe are 'KEYWORD' (which can be
    SWOF, SGOF, etc.), and then SATNUM which is an index counter from 1 and
    onwards. Then follows the data for each individual keyword that
    is found in the deck.

    SATNUM data can only be parsed correctly if TABDIMS is present
    and stating how many saturation functions there should be.
    If you have a string with TABDIMS missing, you must supply
    this as a string to this function, and not a parsed deck, as
    the default parser in EclFiles is very permissive (and only
    returning the first function by default).

    Arguments:
        deck: Incoming data deck. Always
            supply as a string if you don't know TABDIMS-NTSFUN.
        keywords: Requested keywords for which to
            to extract data.
        ntsfun: Number of SATNUMs defined in the deck, only
            needed if TABDIMS with NTSFUN is not found in the deck.
            If not supplied (or None) and NTSFUN is not defined,
            it will be attempted inferred.

    Return:
        pd.DataFrame, columns 'KEYWORD', 'SW', 'KRW', 'KROW', 'PC', ..
    """
    if isinstance(deck, EclFiles):
        # NB: If this is done on include files and not on DATA files
        # we can loose data for SATNUM > 1
        deck = deck.get_ecldeck()
    deck = inferdims.inject_xxxdims_ntxxx("TABDIMS", "NTSFUN", deck, ntsfun)
    assert "TABDIMS" in deck

    wanted_keywords = common.handle_wanted_keywords(keywords, deck, SUPPORTED_KEYWORDS)

    frames = []
    for keyword in wanted_keywords:
        frames.append(
            interpolate_defaults(
                common.ecl_keyworddata_to_df(
                    deck, keyword, renamer=RENAMERS[keyword], recordcountername="SATNUM"
                ).assign(KEYWORD=keyword)
            )
        )
    nonempty_frames = [frame for frame in frames if not frame.empty]
    if nonempty_frames:
        dframe = pd.concat(nonempty_frames, axis=0, sort=False, ignore_index=True)
        # We want to sort the keywords by the order they appear in
        # SUPPORTED_KEYWORDS (mainly to get WaterOil before GasOil)
        # We do that by converting to a Categorical series:
        dframe["KEYWORD"] = pd.Categorical(dframe["KEYWORD"], SUPPORTED_KEYWORDS)
        dframe.sort_values(["SATNUM", "KEYWORD"], inplace=True)
        dframe["KEYWORD"] = dframe["KEYWORD"].astype(str)
        logger.info(
            "Extracted keywords %s for %i SATNUMs",
            dframe["KEYWORD"].unique(),
            len(dframe["SATNUM"].unique()),
        )
        return dframe
    return pd.DataFrame()


def interpolate_defaults(dframe: pd.DataFrame) -> pd.DataFrame:
    """Interpolate NaN's linearly in saturation.
    Saturation function tables in Eclipse decks can have certain values defaulted.
    When parsed by common.ecl2df, these values are returned as np.nan.
    The incoming dataframe must be associated to one keyword only, but
    can consist of multiple SATNUMs.
    """
    sat_cols: set = {"SW", "SO", "SG", "SL"}.intersection(dframe.columns)
    assert (
        len(sat_cols) == 1
    ), f"Could not determine a single saturation column in {dframe.columns}"
    sat_col = list(sat_cols)[0]

    if dframe[sat_col].isna().any():
        raise ValueError("nan in saturation column is not allowed")

    filled_frames = []
    for _, subframe in dframe.groupby("SATNUM"):
        filled_frames.append(
            subframe.set_index(sat_col)
            .interpolate(method="index", limit_area="inside")
            .reset_index()
        )
    return pd.concat(filled_frames)


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser (ArgumentParser or subparser): parser to fill with arguments
    """
    parser.add_argument(
        "DATAFILE", help="Name of Eclipse DATA file or file with saturation functions."
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file.",
        default="satfuncs.csv",
    )
    parser.add_argument(
        "-k",
        "--keywords",
        nargs="+",
        help=(
            "List of saturation function keywords to fetch data from. "
            "If not supplied, all supported keywords will be included."
        ),
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def fill_reverse_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Fill a parser for the operation dataframe -> eclipse include file"""
    return common.fill_reverse_parser(parser, "SWOF, SGOF++", "relperm.inc")


def satfunc_main(args) -> None:
    """Entry-point for module, for command line utility"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    if eclfiles:
        deck = eclfiles.get_ecldeck()
    if "TABDIMS" in deck:
        # Things are easier when a full deck with (correct) TABDIMS
        # is supplied:
        satfunc_df = df(eclfiles, keywords=args.keywords)
    else:
        # This might be an include file for which we have to infer/guess
        # TABDIMS. Then we send it to df() as a string
        satfunc_df = df(
            Path(args.DATAFILE).read_text(encoding="utf-8"), keywords=args.keywords
        )
    if "SATNUM" in satfunc_df and "KEYWORD" in satfunc_df:
        satnums = str(len(satfunc_df["SATNUM"].unique()))
        keywords = str(satfunc_df["KEYWORD"].unique())
    else:
        satnums = "-"
        keywords = "-"
    write_dframe_stdout_file(
        satfunc_df,
        args.output,
        index=False,
        caller_logger=logger,
        logstr=f"Unique SATNUMs: {satnums}, saturation keywords: {keywords}",
    )


def satfunc_reverse_main(args) -> None:
    """For command line utility for CSV to Eclipse"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    satfunc_df = pd.read_csv(args.csvfile)
    logger.info("Parsed %s", args.csvfile)
    inc_string = df2ecl(satfunc_df, keywords=args.keywords)
    common.write_inc_stdout_file(inc_string, args.output)


def df2ecl(
    satfunc_df: pd.DataFrame,
    keywords: Optional[List[str]] = None,
    comments: Optional[Dict[str, str]] = None,
    filename: Optional[str] = None,
) -> str:
    """Generate Eclipse include strings from dataframes with
    saturation functions (SWOF, SGOF, ...)

    Args:
        satfunc_df: Dataframe with data on ecl2df format.
        keywords: List of keywords to include. Must be
            supported and present in the incoming dataframe. Keywords
            are printed in the order defined by this list.
        comments: Dictionary indexed by keyword with comments to be
            included pr. keyword. If a key named "master" is present
            it will be used as a master comment for the outputted file.
        filename: If supplied, the generated text will also be dumped
            to file.

    Returns:
        Generated Eclipse include string

    """
    string = ""
    string += common.df2ecl(
        satfunc_df,
        keywords=keywords,
        comments=comments,
        supported=SUPPORTED_KEYWORDS,
        consecutive="SATNUM",
        filename=filename,
    )
    return string


def df2ecl_swof(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SWOF data. Used by df2ecl().

    Args:
        dframe: Containing SWOF data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SWOF", dframe, comment)


def df2ecl_sgof(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SGOF data. Used by df2ecl().

    Args:
        dframe: Containing SGOF data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SGOF", dframe, comment)


def df2ecl_sgfn(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SGFN data. Used by df2ecl().

    Args:
        dframe: Containing SGFN data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SGFN", dframe, comment)


def df2ecl_sgwfn(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SGWFN data. Used by df2ecl().

    Args:
        dframe: Containing SGWFN data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SGWFN", dframe, comment)


def df2ecl_swfn(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SWFN data. Used by df2ecl().

    Args:
        dframe: Containing SWFN data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SWFN", dframe, comment)


def df2ecl_slgof(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SLGOF data. Used by df2ecl().

    Args:
        dframe: Containing SLGOF data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SLGOF", dframe, comment)


def df2ecl_sof2(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SOF2 data. Used by df2ecl().

    Args:
        dframe: Containing SOF2 data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SOF2", dframe, comment)


def df2ecl_sof3(dframe: pd.DataFrame, comment: Optional[str] = None) -> str:
    """Print SOF3 data. Used by df2ecl().

    Args:
        dframe: Containing SOF3 data
        comment: Text that will be included as a comment
    """
    return _df2ecl_satfuncs("SOF3", dframe, comment)


def _df2ecl_satfuncs(
    keyword: str, dframe: pd.DataFrame, comment: Optional[str] = None
) -> str:
    if dframe.empty:
        return "-- No data!\n"
    string = f"{keyword}\n"
    string += common.comment_formatter(comment)

    if "KEYWORD" not in dframe:
        # Use everything..
        subset = dframe
    else:
        subset = dframe[dframe["KEYWORD"] == keyword]
    if "SATNUM" not in subset:
        subset["SATNUM"] = 1
    subset = subset.set_index("SATNUM").sort_index()

    # Make a function that is to be called for each SATNUM
    def _df2ecl_satfuncs_satnum(keyword, dframe):
        """Print one saturation function for one specific SATNUM"""
        col_headers = RENAMERS[keyword]["DATA"]
        string = (
            "-- "
            + dframe[col_headers]
            .to_string(float_format=" %g", header=True, index=False)
            .strip()
        )
        return string + "\n/\n"

    # Loop over every SATNUM
    for satnum in subset.index.unique():
        string += f"-- SATNUM: {satnum}\n"
        string += _df2ecl_satfuncs_satnum(keyword, subset[subset.index == satnum])
    return string + "\n"



#!/usr/bin/env python
"""
Extract the contents of the FAULTS keyword into
a DataFrame

"""
import argparse
import logging
from typing import Union

import pandas as pd

from ecl2df import EclFiles, getLogger_ecl2csv
from ecl2df.common import parse_opmio_deckrecord, write_dframe_stdout_file

try:
    # Needed for mypy

    # pylint: disable=unused-import
    import opm.io

except ImportError:
    pass


logger = logging.getLogger(__name__)

RECORD_COLUMNS = ["NAME", "IX1", "IX2", "IY1", "IY2", "IZ1", "IZ2", "FACE"]
COLUMNS = ["NAME", "I", "J", "K", "FACE"]
ALLOWED_FACES = ["X", "Y", "Z", "I", "J", "K", "X-", "Y-", "Z-", "I-", "J-", "K-"]


def df(deck: Union[EclFiles, "opm.libopmcommon_python.Deck"]) -> pd.DataFrame:
    """Produce a dataframe of fault data from a deck

    All data for the keyword FAULTS will be returned.

    Args:
        deck: Eclipse deck
    """
    if isinstance(deck, EclFiles):
        deck = deck.get_ecldeck()

    # In[91]: list(deck['FAULTS'][0])
    # Out[91]: [[u'F1'], [36], [36], [41], [42], [1], [14], [u'I']]
    data = []
    # It is allowed in Eclipse to use the keyword FAULTS
    # as many times as needed. Thus we need to loop in some way:
    for keyword in deck:
        if keyword.name == "FAULTS":
            for rec in keyword:
                # Each record now has a range potentially in three
                # dimensions for the fault, unroll this:
                frec_dict = parse_opmio_deckrecord(rec, "FAULTS")
                faultname = frec_dict["NAME"]
                faultface = frec_dict["FACE"]
                for i_idx in range(frec_dict["IX1"], frec_dict["IX2"] + 1):
                    for j_idx in range(frec_dict["IY1"], frec_dict["IY2"] + 1):
                        for k_idx in range(frec_dict["IZ1"], frec_dict["IZ2"] + 1):
                            data.append([faultname, i_idx, j_idx, k_idx, faultface])
    dframe = pd.DataFrame(columns=COLUMNS, data=data)
    logger.info("Extracted %i faults", len(dframe["NAME"].unique()))
    return dframe


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser: argparse.ArgumentParser or argparse.subparser
    """
    parser.add_argument("DATAFILE", help="Name of Eclipse DATA file.")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file.",
        default="faults.csv",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def faults_main(args) -> None:
    """Read from disk and write CSV back to disk"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    if eclfiles:
        deck = eclfiles.get_ecldeck()
    faults_df = df(deck)
    write_dframe_stdout_file(
        faults_df,
        args.output,
        index=False,
        caller_logger=logger,
        logstr=f"Wrote to {args.output}",
    )



import itertools

import pytest

import ecl2df

from .test_grid import EIGHTCELLS, REEK

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False


def test_default_logger_levels_and_split(capsys):
    """Verify that the intended usage of this logger have expected results"""

    splitlogger = ecl2df.getLogger_ecl2csv("test_levels_split")

    splitlogger.debug("This DEBUG-text is not to be seen")
    captured = capsys.readouterr()
    assert "DEBUG-text" not in captured.out
    assert "DEBUG-text" not in captured.err

    splitlogger.info("This INFO-text is not to be seen by default")
    captured = capsys.readouterr()
    assert "INFO-text" not in captured.out

    splitlogger.warning("This WARNING-text is to be seen in stdout")
    captured = capsys.readouterr()
    assert "WARNING-text" in captured.out
    assert "WARNING-text" not in captured.err

    splitlogger.error("This ERROR-text should only be in stderr")
    captured = capsys.readouterr()
    assert "ERROR-text" not in captured.out
    assert "ERROR-text" in captured.err

    # If output is written to stdout, all logs should go to stderr:
    nosplit_logger = ecl2df.getLogger_ecl2csv(
        "test_levels_nosplit", args_dict={"output": "-", "debug": True}
    )
    nosplit_logger.debug("This DEBUG-text is to be seen in stderr")
    captured = capsys.readouterr()
    assert "DEBUG-text" not in captured.out
    assert "DEBUG-text" in captured.err

    nosplit_logger.info("This INFO-text is to be seen by in stderr")
    captured = capsys.readouterr()
    assert "INFO-text" not in captured.out
    assert "INFO-text" in captured.err

    nosplit_logger.warning("This WARNING-text is to be seen in stderr")
    captured = capsys.readouterr()
    assert "WARNING-text" not in captured.out
    assert "WARNING-text" in captured.err

    nosplit_logger.error("This ERROR-text should only be in stderr")
    captured = capsys.readouterr()
    assert "ERROR-text" not in captured.out
    assert "ERROR-text" in captured.err


@pytest.mark.skipif(not HAVE_OPM, reason="Command line client requires OPM")
@pytest.mark.parametrize(
    "ecl2df_module, verbose, fileexport",
    itertools.product(ecl2df.SUBMODULES, [False, True], [True, False]),
)
def test_ecl2csv_logging(tmp_path, ecl2df_module, verbose, fileexport, mocker, capsys):
    """Test that the command line client for each submodule logs correctly.

    Each submodule should write logs to stdout for INFO and WARNING messages
    when they write to dedicated output files, but must write to stderr when
    stdout is used as the CSV output stream. This requres correct configuration
    in each submodule and must therefore be tested.

    This test function is more robust if each main() invocation is run in a
    subprocess, but that also makes it 10 times slower. When not run in a
    subprocess, the verbosity option must be False before True for the tests to
    work, this is related (?) to loggers not being properly reset between each
    test invocation.
    """

    if "nnc" == ecl2df_module:
        # There are no nnc's in EIGHTCELLS, so for that test
        # we need the REEK dataset:
        commands = ["ecl2csv", ecl2df_module, REEK, "--output"]
    else:
        commands = ["ecl2csv", ecl2df_module, EIGHTCELLS, "--output"]

    if fileexport:
        commands.append(str(tmp_path / "output.csv"))
    else:
        commands.append(ecl2df.common.MAGIC_STDOUT)

    if verbose:
        commands.append("-v")

    mocker.patch("sys.argv", commands)

    ecl2df.ecl2csv.main()
    captured = capsys.readouterr()
    stdout_output = captured.out
    stderr_output = captured.err

    if fileexport:
        if verbose:
            assert "INFO:" in stdout_output
            assert "INFO:" not in stderr_output
        else:
            assert "INFO:" not in stdout_output
            assert "INFO:" not in stderr_output
    else:  # "CSV" data is dumped to stdout
        if verbose:
            assert "INFO:" in stderr_output
            assert "INFO:" not in stdout_output
        else:
            assert "INFO:" not in stdout_output
            assert "INFO:" not in stderr_output


def test_repeated_logger_construction(capsys):
    """If we repeatedly call getLogger(), ensure handlers are not added on top"""
    logger = ecl2df.getLogger_ecl2csv("nodouble")
    logger = ecl2df.getLogger_ecl2csv("nodouble")
    logger.warning("Don't repeat me")
    captured = capsys.readouterr()
    assert captured.out.count("Don't repeat me") == 1



import importlib
import logging
import sys
from typing import Dict, List, Union

try:
    from .version import version  # type: ignore

    __version__ = version
except ImportError:
    __version__ = "v0.0.0"

from .constants import MAGIC_STDOUT
from .eclfiles import EclFiles

SUBMODULES: List[str] = [
    "compdat",
    "equil",
    "faults",
    "fipreports",
    "grid",
    "gruptree",
    "nnc",
    "pillars",
    "pvt",
    "rft",
    "satfunc",
    "summary",
    "trans",
    "wellconnstatus",
    "wcon",
]


def getLogger_ecl2csv(
    module_name: str = "ecl2df", args_dict: Dict[str, Union[str, bool]] = None
) -> logging.Logger:
    # pylint: disable=invalid-name
    """Provide a custom logger for ecl2csv and csv2ecl

    Logging output is by default split by logging levels (split between WARNING and
    ERROR) to stdout and stderr, each log occurs in only one of the streams.
    This deviates from Unix standard, but is accepted here because the code
    often writes it main output to a dedicated filename.

    Args:
        module_name: A suggested name for the logger, usually
            __name__ should be supplied
        args_dict: Dictionary with contents from the argparse namespace object.
            Only keys "output", "verbose" and "debug" will be looked at.
            Use vars(args) in caller code to convert from argparse Namespace object.
    """
    logger = logging.getLogger(module_name)

    if args_dict is None:
        args_dict = {}

    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")

    # In case this logger has been initialized earlier, clear it:
    logger.handlers.clear()

    if args_dict.get("output", "") == MAGIC_STDOUT:
        # If main output is to stdout, we must send all logs to stderr:
        default_handler = logging.StreamHandler(sys.stderr)
        default_handler.setFormatter(formatter)
        logger.addHandler(default_handler)
    else:
        # Split log messages to either stdout or stderr based on log level:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.addFilter(lambda record: record.levelno < logging.ERROR)
        stdout_handler.setFormatter(formatter)

        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.addFilter(lambda record: record.levelno >= logging.ERROR)
        stderr_handler.setFormatter(formatter)

        logger.addHandler(stdout_handler)
        logger.addHandler(stderr_handler)

    # --debug overrides --verbose
    if args_dict.get("debug", False):
        logger.setLevel(logging.DEBUG)
    elif args_dict.get("verbose", False):
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    return logger


for submodule in SUBMODULES + ["ecl2csv", "csv2ecl"]:
    importlib.import_module("ecl2df." + submodule)
